# -*- coding: utf-8 -*-

import re

from .. import EloquentTestCase
from .. import mock

from ..utils import MockConnection, MockProcessor

from eloquent.query.grammars import (
    QueryGrammar,
    PostgresQueryGrammar,
    SQLiteQueryGrammar,
    MySqlQueryGrammar
)
from eloquent.query.builder import QueryBuilder
from eloquent.query.expression import QueryExpression
from eloquent.query.join_clause import JoinClause


class QueryBuilderTestCase(EloquentTestCase):

    def test_basic_select(self):
        builder = self.get_builder()
        builder.select('*').from_('users')

        self.assertEqual('SELECT * FROM "users"', builder.to_sql())

    def test_basic_select_use_write_connection(self):
        builder = self.get_builder()
        builder.use_write_connection().select('*').from_('users').get()
        builder.get_connection().select.assert_called_once_with(
            'SELECT * FROM "users"',
            [],
            False
        )

        builder = self.get_builder()
        builder.select('*').from_('users').get()
        builder.get_connection().select.assert_called_once_with(
            'SELECT * FROM "users"',
            [],
            True
        )

    def test_alias_wrapping_as_whole_constant(self):
        builder = self.get_builder()
        builder.select('x.y as foo.bar').from_('baz')

        self.assertEqual(
            'SELECT "x"."y" AS "foo.bar" FROM "baz"',
            builder.to_sql()
        )

    def test_adding_selects(self):
        builder = self.get_builder()
        builder.select('foo').add_select('bar').add_select('baz', 'boom').from_('users')

        self.assertEqual(
            'SELECT "foo", "bar", "baz", "boom" FROM "users"',
            builder.to_sql()
        )

    def test_basic_select_with_prefix(self):
        builder = self.get_builder()
        builder.get_grammar().set_table_prefix('prefix_')
        builder.select('*').from_('users')

        self.assertEqual(
            'SELECT * FROM "prefix_users"',
            builder.to_sql()
        )

    def test_basic_select_distinct(self):
        builder = self.get_builder()
        builder.distinct().select('foo', 'bar').from_('users')

        self.assertEqual(
            'SELECT DISTINCT "foo", "bar" FROM "users"',
            builder.to_sql()
        )

    def test_basic_alias(self):
        builder = self.get_builder()
        builder.select('foo as bar').from_('users')

        self.assertEqual(
            'SELECT "foo" AS "bar" FROM "users"',
            builder.to_sql()
        )

    def test_mysql_wrapping_protects_qutotation_marks(self):
        builder = self.get_mysql_builder()
        builder.select('*').from_('some`table')

        self.assertEqual(
            'SELECT * FROM `some``table`',
            builder.to_sql()
        )

    def test_where_day_mysql(self):
        builder = self.get_mysql_builder()
        builder.select('*').from_('users').where_day('created_at', '=', 1)

        self.assertEqual(
            'SELECT * FROM `users` WHERE DAY(`created_at`) = %s',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_where_month_mysql(self):
        builder = self.get_mysql_builder()
        builder.select('*').from_('users').where_month('created_at', '=', 5)

        self.assertEqual(
            'SELECT * FROM `users` WHERE MONTH(`created_at`) = %s',
            builder.to_sql()
        )
        self.assertEqual([5], builder.get_bindings())

    def test_where_year_mysql(self):
        builder = self.get_mysql_builder()
        builder.select('*').from_('users').where_year('created_at', '=', 2014)

        self.assertEqual(
            'SELECT * FROM `users` WHERE YEAR(`created_at`) = %s',
            builder.to_sql()
        )
        self.assertEqual([2014], builder.get_bindings())

    def test_where_day_postgres(self):
        builder = self.get_postgres_builder()
        builder.select('*').from_('users').where_day('created_at', '=', 1)

        self.assertEqual(
            'SELECT * FROM "users" WHERE DAY("created_at") = %s',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_where_month_postgres(self):
        builder = self.get_postgres_builder()
        builder.select('*').from_('users').where_month('created_at', '=', 5)

        self.assertEqual(
            'SELECT * FROM "users" WHERE MONTH("created_at") = %s',
            builder.to_sql()
        )
        self.assertEqual([5], builder.get_bindings())

    def test_where_year_postgres(self):
        builder = self.get_postgres_builder()
        builder.select('*').from_('users').where_year('created_at', '=', 2014)

        self.assertEqual(
            'SELECT * FROM "users" WHERE YEAR("created_at") = %s',
            builder.to_sql()
        )
        self.assertEqual([2014], builder.get_bindings())

    def test_where_day_sqlite(self):
        builder = self.get_sqlite_builder()
        builder.select('*').from_('users').where_day('created_at', '=', 1)

        self.assertEqual(
            'SELECT * FROM "users" WHERE strftime(\'%d\', "created_at") = ?',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_where_month_sqlite(self):
        builder = self.get_sqlite_builder()
        builder.select('*').from_('users').where_month('created_at', '=', 5)

        self.assertEqual(
            'SELECT * FROM "users" WHERE strftime(\'%m\', "created_at") = ?',
            builder.to_sql()
        )
        self.assertEqual([5], builder.get_bindings())

    def test_where_year_sqlite(self):
        builder = self.get_sqlite_builder()
        builder.select('*').from_('users').where_year('created_at', '=', 2014)

        self.assertEqual(
            'SELECT * FROM "users" WHERE strftime(\'%Y\', "created_at") = ?',
            builder.to_sql()
        )
        self.assertEqual([2014], builder.get_bindings())

    def test_where_between(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_between('id', [1, 2])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" BETWEEN ? AND ?',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where_not_between('id', [1, 2])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" NOT BETWEEN ? AND ?',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

    def test_basic_or_where(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where('email', '=', 'foo')
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR "email" = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 'foo'], builder.get_bindings())

    def test_raw_wheres(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_raw('id = ? or email = ?', [1, 'foo'])
        self.assertEqual(
            'SELECT * FROM "users" WHERE id = ? OR email = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 'foo'], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where_raw('id = ?', [1])
        self.assertEqual(
            'SELECT * FROM "users" WHERE id = ?',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_raw_or_wheres(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where_raw('email = ?', ['foo'])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR email = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 'foo'], builder.get_bindings())

    def test_basic_where_ins(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_in('id', [1, 2, 3])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" IN (?, ?, ?)',
            builder.to_sql()
        )
        self.assertEqual([1, 2, 3], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where_in('id', [1, 2, 3])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR "id" IN (?, ?, ?)',
            builder.to_sql()
        )
        self.assertEqual([1, 1, 2, 3], builder.get_bindings())

    def test_basic_where_not_ins(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_not_in('id', [1, 2, 3])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" NOT IN (?, ?, ?)',
            builder.to_sql()
        )
        self.assertEqual([1, 2, 3], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where_not_in('id', [1, 2, 3])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR "id" NOT IN (?, ?, ?)',
            builder.to_sql()
        )
        self.assertEqual([1, 1, 2, 3], builder.get_bindings())

    def test_empty_where_ins(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_in('id', [])
        self.assertEqual(
            'SELECT * FROM "users" WHERE 0 = 1',
            builder.to_sql()
        )
        self.assertEqual([], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where_in('id', [])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR 0 = 1',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_empty_where_not_ins(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_not_in('id', [])
        self.assertEqual(
            'SELECT * FROM "users" WHERE 1 = 1',
            builder.to_sql()
        )
        self.assertEqual([], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where_not_in('id', [])
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR 1 = 1',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_unions(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union(self.get_builder().select('*').from_('users').where('id', '=', 2))
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? UNION SELECT * FROM "users" WHERE "id" = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

        builder = self.get_mysql_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union(self.get_mysql_builder().select('*').from_('users').where('id', '=', 2))
        self.assertEqual(
            '(SELECT * FROM `users` WHERE `id` = %s) UNION (SELECT * FROM `users` WHERE `id` = %s)',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

    def test_union_alls(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union_all(self.get_builder().select('*').from_('users').where('id', '=', 2))
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? UNION ALL SELECT * FROM "users" WHERE "id" = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

    def test_multiple_unions(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union(self.get_builder().select('*').from_('users').where('id', '=', 2))
        builder.union(self.get_builder().select('*').from_('users').where('id', '=', 3))
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? '
            'UNION SELECT * FROM "users" WHERE "id" = ? '
            'UNION SELECT * FROM "users" WHERE "id" = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 2, 3], builder.get_bindings())

    def test_multiple_union_alls(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union_all(self.get_builder().select('*').from_('users').where('id', '=', 2))
        builder.union_all(self.get_builder().select('*').from_('users').where('id', '=', 3))
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? '
            'UNION ALL SELECT * FROM "users" WHERE "id" = ? '
            'UNION ALL SELECT * FROM "users" WHERE "id" = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 2, 3], builder.get_bindings())

    def test_union_order_bys(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union(self.get_builder().select('*').from_('users').where('id', '=', 2))
        builder.order_by('id', 'desc')
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? '
            'UNION SELECT * FROM "users" WHERE "id" = ? '
            'ORDER BY "id" DESC',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

    def test_union_limits_and_offsets(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union(self.get_builder().select('*').from_('users').where('id', '=', 2))
        builder.skip(5).take(10)
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? '
            'UNION SELECT * FROM "users" WHERE "id" = ? '
            'LIMIT 10 OFFSET 5',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

    def test_mysql_union_order_bys(self):
        builder = self.get_mysql_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union(self.get_mysql_builder().select('*').from_('users').where('id', '=', 2))
        builder.order_by('id', 'desc')
        self.assertEqual(
            '(SELECT * FROM `users` WHERE `id` = %s) '
            'UNION (SELECT * FROM `users` WHERE `id` = %s) '
            'ORDER BY `id` DESC',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

    def test_mysql_union_limits_and_offsets(self):
        builder = self.get_mysql_builder()
        builder.select('*').from_('users').where('id', '=', 1)
        builder.union(self.get_mysql_builder().select('*').from_('users').where('id', '=', 2))
        builder.skip(5).take(10)
        self.assertEqual(
            '(SELECT * FROM `users` WHERE `id` = %s) '
            'UNION (SELECT * FROM `users` WHERE `id` = %s) '
            'LIMIT 10 OFFSET 5',
            builder.to_sql()
        )
        self.assertEqual([1, 2], builder.get_bindings())

    def test_sub_select_where_in(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_in(
            'id',
            self.get_builder().select('id').from_('users').where('age', '>', 25).take(3)
        )
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" IN (SELECT "id" FROM "users" WHERE "age" > ? LIMIT 3)',
            builder.to_sql()
        )
        self.assertEqual([25], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where_not_in(
            'id',
            self.get_builder().select('id').from_('users').where('age', '>', 25).take(3)
        )
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" NOT IN (SELECT "id" FROM "users" WHERE "age" > ? LIMIT 3)',
            builder.to_sql()
        )
        self.assertEqual([25], builder.get_bindings())

    def test_basic_where_null(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_null('id')
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" IS NULL',
            builder.to_sql()
        )
        self.assertEqual([], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where_null('id')
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR "id" IS NULL',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_basic_where_not_null(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where_not_null('id')
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" IS NOT NULL',
            builder.to_sql()
        )
        self.assertEqual([], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').where('id', '=', 1).or_where_not_null('id')
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR "id" IS NOT NULL',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_group_bys(self):
        builder = self.get_builder()
        builder.select('*').from_('users').group_by('id', 'email')
        self.assertEqual(
            'SELECT * FROM "users" GROUP BY "id", "email"',
            builder.to_sql()
        )

    def test_order_bys(self):
        builder = self.get_builder()
        builder.select('*').from_('users').order_by('email').order_by('age', 'desc')
        self.assertEqual(
            'SELECT * FROM "users" ORDER BY "email" ASC, "age" DESC',
            builder.to_sql()
        )

        builder = self.get_builder()
        builder.select('*').from_('users').order_by('email').order_by_raw('"age" ? desc', ['foo'])
        self.assertEqual(
            'SELECT * FROM "users" ORDER BY "email" ASC, "age" ? DESC',
            builder.to_sql()
        )

    def test_havings(self):
        builder = self.get_builder()
        builder.select('*').from_('users').having('email', '>', 1)
        self.assertEqual(
            'SELECT * FROM "users" HAVING "email" > ?',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users')\
            .or_having('email', '=', 'foo@bar.com')\
            .or_having('email', '=', 'foo2@bar.com')
        self.assertEqual(
            'SELECT * FROM "users" HAVING "email" = ? OR "email" = ?',
            builder.to_sql()
        )
        self.assertEqual(['foo@bar.com', 'foo2@bar.com'], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users')\
            .group_by('email')\
            .having('email', '>', 1)
        self.assertEqual(
            'SELECT * FROM "users" GROUP BY "email" HAVING "email" > ?',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

        builder = self.get_builder()
        builder.select('email as foo_mail').from_('users')\
            .having('foo_mail', '>', 1)
        self.assertEqual(
            'SELECT "email" AS "foo_mail" FROM "users" HAVING "foo_mail" > ?',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

        builder = self.get_builder()
        builder.select('category', QueryExpression('count(*) as "total"'))\
            .from_('item')\
            .where('department', '=', 'popular')\
            .group_by('category')\
            .having('total', '>', QueryExpression('3'))
        self.assertEqual(
            'SELECT "category", count(*) as "total" '
            'FROM "item" '
            'WHERE "department" = ? '
            'GROUP BY "category" '
            'HAVING "total" > 3',
            builder.to_sql()
        )
        self.assertEqual(['popular'], builder.get_bindings())

        builder = self.get_builder()
        builder.select('category', QueryExpression('count(*) as "total"'))\
            .from_('item')\
            .where('department', '=', 'popular')\
            .group_by('category')\
            .having('total', '>', 3)
        self.assertEqual(
            'SELECT "category", count(*) as "total" '
            'FROM "item" '
            'WHERE "department" = ? '
            'GROUP BY "category" '
            'HAVING "total" > ?',
            builder.to_sql()
        )
        self.assertEqual(['popular', 3], builder.get_bindings())

    def test_having_followed_by_select_get(self):
        builder = self.get_builder()
        query = 'SELECT "category", count(*) as "total" ' \
                'FROM "item" ' \
                'WHERE "department" = ? ' \
                'GROUP BY "category" ' \
                'HAVING "total" > ?'
        results = {
            'category': 'rock',
            'total': 5
        }
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results: results)
        result = builder.select('category', QueryExpression('count(*) as "total"'))\
            .from_('item')\
            .where('department', '=', 'popular')\
            .group_by('category')\
            .having('total', '>', 3)\
            .get()

        builder.get_connection().select.assert_called_once_with(
            query,
            ['popular', 3],
            True
        )
        builder.get_processor().process_select.assert_called_once_with(
            builder,
            results
        )
        self.assertEqual(results, result)
        self.assertEqual(['popular', 3], builder.get_bindings())

        # Using raw value
        builder = self.get_builder()
        query = 'SELECT "category", count(*) as "total" ' \
                'FROM "item" ' \
                'WHERE "department" = ? ' \
                'GROUP BY "category" ' \
                'HAVING "total" > 3'
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results: results)
        result = builder.select('category', QueryExpression('count(*) as "total"'))\
            .from_('item')\
            .where('department', '=', 'popular')\
            .group_by('category')\
            .having('total', '>', QueryExpression('3'))\
            .get()

        builder.get_connection().select.assert_called_once_with(
            query,
            ['popular'],
            True
        )
        builder.get_processor().process_select.assert_called_once_with(
            builder,
            results
        )
        self.assertEqual(results, result)
        self.assertEqual(['popular'], builder.get_bindings())

    def test_raw_havings(self):
        builder = self.get_builder()
        builder.select('*').from_('users').having_raw('user_foo < user_bar')
        self.assertEqual(
            'SELECT * FROM "users" HAVING user_foo < user_bar',
            builder.to_sql()
        )
        self.assertEqual([], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').having('foo', '=', 1).or_having_raw('user_foo < user_bar')
        self.assertEqual(
            'SELECT * FROM "users" HAVING "foo" = ? OR user_foo < user_bar',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_limits_and_offsets(self):
        builder = self.get_builder()
        builder.select('*').from_('users').offset(5).limit(10)
        self.assertEqual('SELECT * FROM "users" LIMIT 10 OFFSET 5', builder.to_sql())

        builder = self.get_builder()
        builder.select('*').from_('users').skip(5).take(10)
        self.assertEqual('SELECT * FROM "users" LIMIT 10 OFFSET 5', builder.to_sql())

        builder = self.get_builder()
        builder.select('*').from_('users').skip(-5).take(10)
        self.assertEqual('SELECT * FROM "users" LIMIT 10 OFFSET 0', builder.to_sql())

        builder = self.get_builder()
        builder.select('*').from_('users').for_page(2, 15)
        self.assertEqual('SELECT * FROM "users" LIMIT 15 OFFSET 15', builder.to_sql())

        builder = self.get_builder()
        builder.select('*').from_('users').for_page(-2, 15)
        self.assertEqual('SELECT * FROM "users" LIMIT 15 OFFSET 0', builder.to_sql())

    def test_where_shortcut(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('id', 1).or_where('name', 'foo')
        self.assertEqual(
            'SELECT * FROM "users" WHERE "id" = ? OR "name" = ?',
            builder.to_sql()
        )
        self.assertEqual([1, 'foo'], builder.get_bindings())

    def test_nested_wheres(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('email', '=', 'foo').or_where(
            builder.new_query().where('name', '=', 'bar').where('age', '=', 25)
        )
        self.assertEqual(
            'SELECT * FROM "users" WHERE "email" = ? OR ("name" = ? AND "age" = ?)',
            builder.to_sql()
        )
        self.assertEqual(['foo', 'bar', 25], builder.get_bindings())

    def test_full_sub_selects(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('email', '=', 'foo').or_where(
            'id', '=', builder.new_query().select(QueryExpression('max(id)')).from_('users').where('email', '=', 'bar')
        )
        self.assertEqual(
            'SELECT * FROM "users" WHERE "email" = ? OR "id" = (SELECT max(id) FROM "users" WHERE "email" = ?)',
            builder.to_sql()
        )
        self.assertEqual(['foo', 'bar'], builder.get_bindings())

    def test_where_exists(self):
        builder = self.get_builder()
        builder.select('*').from_('orders').where_exists(
            self.get_builder().select('*').from_('products').where('products.id', '=', QueryExpression('"orders"."id"'))
        )
        self.assertEqual(
            'SELECT * FROM "orders" '
            'WHERE EXISTS (SELECT * FROM "products" WHERE "products"."id" = "orders"."id")',
            builder.to_sql()
        )
        self.assertEqual([], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('orders').where_not_exists(
            self.get_builder().select('*').from_('products').where('products.id', '=', QueryExpression('"orders"."id"'))
        )
        self.assertEqual(
            'SELECT * FROM "orders" '
            'WHERE NOT EXISTS (SELECT * FROM "products" WHERE "products"."id" = "orders"."id")',
            builder.to_sql()
        )
        self.assertEqual([], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('orders').where('id', '=', 1).or_where_exists(
            self.get_builder().select('*').from_('products').where('products.id', '=', QueryExpression('"orders"."id"'))
        )
        self.assertEqual(
            'SELECT * FROM "orders" WHERE "id" = ? '
            'OR EXISTS (SELECT * FROM "products" WHERE "products"."id" = "orders"."id")',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('orders').where('id', '=', 1).or_where_not_exists(
            self.get_builder().select('*').from_('products').where('products.id', '=', QueryExpression('"orders"."id"'))
        )
        self.assertEqual(
            'SELECT * FROM "orders" WHERE "id" = ? '
            'OR NOT EXISTS (SELECT * FROM "products" WHERE "products"."id" = "orders"."id")',
            builder.to_sql()
        )
        self.assertEqual([1], builder.get_bindings())

    def test_basic_joins(self):
        builder = self.get_builder()
        builder.select('*').from_('users')\
            .join('contacts', 'users.id', '=', 'contacts.id')\
            .left_join('photos', 'users.id', '=', 'photos.user_id')
        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" ON "users"."id" = "contacts"."id" '
            'LEFT JOIN "photos" ON "users"."id" = "photos"."user_id"',
            builder.to_sql()
        )

        builder = self.get_builder()
        builder.select('*').from_('users')\
            .left_join_where('photos', 'users.id', '=', 3)\
            .join_where('photos', 'users.id', '=', 'foo')
        self.assertEqual(
            'SELECT * FROM "users" '
            'LEFT JOIN "photos" ON "users"."id" = ? '
            'INNER JOIN "photos" ON "users"."id" = ?',
            builder.to_sql()
        )
        self.assertEqual([3, 'foo'], builder.get_bindings())

    def test_complex_joins(self):
        builder = self.get_builder()
        builder.select('*').from_('users').join(
            JoinClause('contacts')
            .on('users.id', '=', 'contacts.id')
            .or_on('users.name', '=', 'contacts.name')
        )
        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" ON "users"."id" = "contacts"."id" '
            'OR "users"."name" = "contacts"."name"',
            builder.to_sql()
        )

        builder = self.get_builder()
        builder.select('*').from_('users').join(
            JoinClause('contacts')
            .where('users.id', '=', 'foo')
            .or_where('users.name', '=', 'bar')
        )
        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" ON "users"."id" = ? '
            'OR "users"."name" = ?',
            builder.to_sql()
        )
        self.assertEqual(['foo', 'bar'], builder.get_bindings())

        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" ON "users"."id" = ? '
            'OR "users"."name" = ?',
            builder.to_sql()
        )
        self.assertEqual(['foo', 'bar'], builder.get_bindings())

        builder = self.get_builder()
        builder.select('*').from_('users').left_join(
            JoinClause('contacts')
            .where('users.id', '=', 'foo')
            .or_where('users.name', '=', 'bar')
        )
        self.assertEqual(
            'SELECT * FROM "users" '
            'LEFT JOIN "contacts" ON "users"."id" = ? '
            'OR "users"."name" = ?',
            builder.to_sql()
        )
        self.assertEqual(['foo', 'bar'], builder.get_bindings())

        self.assertEqual(
            'SELECT * FROM "users" '
            'LEFT JOIN "contacts" ON "users"."id" = ? '
            'OR "users"."name" = ?',
            builder.to_sql()
        )
        self.assertEqual(['foo', 'bar'], builder.get_bindings())

    def test_join_where_null(self):
        builder = self.get_builder()
        builder.select('*').from_('users').join(
            JoinClause('contacts')
            .on('users.id', '=', 'contacts.id')
            .where_null('contacts.deleted_at')
        )
        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" '
            'ON "users"."id" = "contacts"."id" '
            'AND "contacts"."deleted_at" IS NULL',
            builder.to_sql()
        )

        builder = self.get_builder()
        builder.select('*').from_('users').join(
            JoinClause('contacts')
            .on('users.id', '=', 'contacts.id')
            .or_where_null('contacts.deleted_at')
        )
        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" '
            'ON "users"."id" = "contacts"."id" '
            'OR "contacts"."deleted_at" IS NULL',
            builder.to_sql()
        )

    def test_join_where_not_null(self):
        builder = self.get_builder()
        builder.select('*').from_('users').join(
            JoinClause('contacts')
            .on('users.id', '=', 'contacts.id')
            .where_not_null('contacts.deleted_at')
        )
        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" '
            'ON "users"."id" = "contacts"."id" '
            'AND "contacts"."deleted_at" IS NOT NULL',
            builder.to_sql()
        )

        builder = self.get_builder()
        builder.select('*').from_('users').join(
            JoinClause('contacts')
            .on('users.id', '=', 'contacts.id')
            .or_where_not_null('contacts.deleted_at')
        )
        self.assertEqual(
            'SELECT * FROM "users" '
            'INNER JOIN "contacts" '
            'ON "users"."id" = "contacts"."id" '
            'OR "contacts"."deleted_at" IS NOT NULL',
            builder.to_sql()
        )

    def test_raw_expression_in_select(self):
        builder = self.get_builder()
        builder.select(QueryExpression('substr(foo, 6)')).from_('users')
        self.assertEqual('SELECT substr(foo, 6) FROM "users"', builder.to_sql())

    def test_find_return_first_result_by_id(self):
        builder = self.get_builder()
        query = 'SELECT * FROM "users" WHERE "id" = ? LIMIT 1'
        results = [{
            'foo': 'bar'
        }]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').find(1)
        builder.get_connection().select.assert_called_once_with(
            query, [1], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual(result, results[0])

    def test_first_return_first_result(self):
        builder = self.get_builder()
        query = 'SELECT * FROM "users" WHERE "id" = ? LIMIT 1'
        results = [{
            'foo': 'bar'
        }]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').where('id', '=', 1).first()
        builder.get_connection().select.assert_called_once_with(
            query, [1], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual(result, results[0])

    def test_list_methods_gets_list_of_colmun_values(self):
        builder = self.get_builder()
        results = [
            {'foo': 'bar'}, {'foo': 'baz'}
        ]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').where('id', '=', 1).lists('foo')
        self.assertEqual(['bar', 'baz'], result)

        builder = self.get_builder()
        results = [
            {'id': 1, 'foo': 'bar'}, {'id': 10, 'foo': 'baz'}
        ]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').where('id', '=', 1).lists('foo', 'id')
        self.assertEqual({1: 'bar', 10: 'baz'}, result)

    def test_implode(self):
        builder = self.get_builder()
        results = [
            {'foo': 'bar'}, {'foo': 'baz'}
        ]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').where('id', '=', 1).implode('foo')
        self.assertEqual('barbaz', result)

        builder = self.get_builder()
        results = [
            {'foo': 'bar'}, {'foo': 'baz'}
        ]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').where('id', '=', 1).implode('foo', ',')
        self.assertEqual('bar,baz', result)

    def test_pluck_return_single_column(self):
        builder = self.get_builder()
        query = 'SELECT "foo" FROM "users" WHERE "id" = ? LIMIT 1'
        results = [{'foo': 'bar'}]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').where('id', '=', 1).pluck('foo')
        builder.get_connection().select.assert_called_once_with(
            query, [1], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual('bar', result)

    def test_aggegate_functions(self):
        builder = self.get_builder()
        query = 'SELECT COUNT(*) AS aggregate FROM "users"'
        results = [{'aggregate': 1}]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results)
        result = builder.from_('users').count()
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual(1, result)

        builder = self.get_builder()
        query = 'SELECT COUNT(*) AS aggregate FROM "users" LIMIT 1'
        results = [{'aggregate': 1}]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        result = builder.from_('users').exists()
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertTrue(result)

        builder = self.get_builder()
        query = 'SELECT MAX("id") AS aggregate FROM "users"'
        results = [{'aggregate': 1}]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        result = builder.from_('users').max('id')
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual(1, result)

        builder = self.get_builder()
        query = 'SELECT MIN("id") AS aggregate FROM "users"'
        results = [{'aggregate': 1}]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        result = builder.from_('users').min('id')
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual(1, result)

        builder = self.get_builder()
        query = 'SELECT SUM("id") AS aggregate FROM "users"'
        results = [{'aggregate': 1}]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        result = builder.from_('users').sum('id')
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual(1, result)

        builder = self.get_builder()
        query = 'SELECT AVG("id") AS aggregate FROM "users"'
        results = [{'aggregate': 1}]
        builder.get_connection().select.return_value = results
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        result = builder.from_('users').avg('id')
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, results)
        self.assertEqual(1, result)

    def test_aggregate_reset_followed_by_get(self):
        builder = self.get_builder()
        query = 'SELECT COUNT(*) AS aggregate FROM "users"'
        builder.get_connection().select.return_value = [{'aggregate': 1}]
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        builder.from_('users').select('column1', 'column2')
        count = builder.count()
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, [{'aggregate': 1}])
        self.assertEqual(1, count)

        builder.get_connection().select.reset_mock()
        builder.get_processor().process_select.reset_mock()
        query = 'SELECT SUM("id") AS aggregate FROM "users"'
        builder.get_connection().select.return_value = [{'aggregate': 2}]
        sum_ = builder.sum('id')
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, [{'aggregate': 2}])
        self.assertEqual(2, sum_)

        builder.get_connection().select.reset_mock()
        builder.get_processor().process_select.reset_mock()
        query = 'SELECT "column1", "column2" FROM "users"'
        builder.get_connection().select.return_value = [{'column1': 'foo', 'column2': 'bar'}]
        result = builder.get()
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, [{'column1': 'foo', 'column2': 'bar'}])
        self.assertEqual([{'column1': 'foo', 'column2': 'bar'}], result)

    def test_aggregate_reset_followed_by_select_get(self):
        builder = self.get_builder()
        query = 'SELECT COUNT("column1") AS aggregate FROM "users"'
        builder.get_connection().select.return_value = [{'aggregate': 1}]
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        builder.from_('users')
        count = builder.count('column1')
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, [{'aggregate': 1}])
        self.assertEqual(1, count)

        builder.get_connection().select.reset_mock()
        builder.get_processor().process_select.reset_mock()
        query = 'SELECT "column2", "column3" FROM "users"'
        builder.get_connection().select.return_value = [{'column2': 'foo', 'column3': 'bar'}]
        result = builder.select('column2', 'column3').get()
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, [{'column2': 'foo', 'column3': 'bar'}])
        self.assertEqual([{'column2': 'foo', 'column3': 'bar'}], result)

    def test_aggregate_reset_followed_by_get_with_columns(self):
        builder = self.get_builder()
        query = 'SELECT COUNT("column1") AS aggregate FROM "users"'
        builder.get_connection().select.return_value = [{'aggregate': 1}]
        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)
        builder.from_('users')
        count = builder.count('column1')
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, [{'aggregate': 1}])
        self.assertEqual(1, count)

        builder.get_connection().select.reset_mock()
        builder.get_processor().process_select.reset_mock()
        query = 'SELECT "column2", "column3" FROM "users"'
        builder.get_connection().select.return_value = [{'column2': 'foo', 'column3': 'bar'}]
        result = builder.get(['column2', 'column3'])
        builder.get_connection().select.assert_called_once_with(
            query, [], True
        )
        builder.get_processor().process_select.assert_called_once_with(builder, [{'column2': 'foo', 'column3': 'bar'}])
        self.assertEqual([{'column2': 'foo', 'column3': 'bar'}], result)

    def test_insert_method(self):
        builder = self.get_builder()
        query = 'INSERT INTO "users" ("email") VALUES (?)'
        builder.get_connection().insert.return_value = True
        result = builder.from_('users').insert({'email': 'foo'})
        builder.get_connection().insert.assert_called_once_with(
            query, ['foo']
        )
        self.assertTrue(result)

    def test_insert_method_with_keyword_arguments(self):
        builder = self.get_builder()
        query = 'INSERT INTO "users" ("email") VALUES (?)'
        builder.get_connection().insert.return_value = True
        result = builder.from_('users').insert({'email': 'foo'})
        builder.get_connection().insert.assert_called_once_with(
            query, ['foo']
        )
        self.assertTrue(result)

    def test_sqlite_multiple_insert(self):
        builder = self.get_sqlite_builder()
        query = 'INSERT INTO "users" ("email", "name") ' \
                'SELECT ? AS "email", ? AS "name" UNION ALL SELECT ? AS "email", ? AS "name"'
        builder.get_connection().insert.return_value = True
        result = builder.from_('users').insert([
            {'email': 'foo', 'name': 'john'},
            {'email': 'bar', 'name': 'jane'}
        ])
        builder.get_connection().insert.assert_called_once_with(
            query, ['foo', 'john', 'bar', 'jane']
        )
        self.assertTrue(result)

    def test_insert_get_id_method(self):
        builder = self.get_builder()
        builder.get_processor().process_insert_get_id.return_value = 1
        result = builder.from_('users').insert_get_id({
            'email': 'foo',
            'bar': QueryExpression('bar')
        })
        builder.get_processor().process_insert_get_id.assert_called_once_with(
            builder, 'INSERT INTO "users" ("bar", "email") VALUES (bar, ?)', ['foo'], None
        )
        self.assertEqual(1, result)

    def test_insert_get_id_with_sequence(self):
        builder = self.get_builder()
        builder.get_processor().process_insert_get_id.return_value = 1
        result = builder.from_('users').insert_get_id({
            'email': 'foo',
            'bar': QueryExpression('bar')
        }, 'id')
        builder.get_processor().process_insert_get_id.assert_called_once_with(
            builder, 'INSERT INTO "users" ("bar", "email") VALUES (bar, ?)', ['foo'], 'id'
        )
        self.assertEqual(1, result)

    def test_insert_get_id_respects_raw_bindings(self):
        builder = self.get_builder()
        builder.get_processor().process_insert_get_id.return_value = 1
        result = builder.from_('users').insert_get_id({
            'email': QueryExpression('CURRENT_TIMESTAMP'),
        })
        builder.get_processor().process_insert_get_id.assert_called_once_with(
            builder, 'INSERT INTO "users" ("email") VALUES (CURRENT_TIMESTAMP)', [], None
        )
        self.assertEqual(1, result)

    def test_update(self):
        builder = self.get_builder()
        query = 'UPDATE "users" SET "email" = ?, "name" = ? WHERE "id" = ?'
        builder.get_connection().update.return_value = 1
        result = builder.from_('users').where('id', '=', 1).update(email='foo', name='bar')
        builder.get_connection().update.assert_called_with(
            query, ['foo', 'bar', 1]
        )
        self.assertEqual(1, result)
        
        builder = self.get_mysql_builder()
        marker = builder.get_grammar().get_marker()
        query = 'UPDATE `users` SET `email` = %s, `name` = %s WHERE `id` = %s' % (marker, marker, marker)
        builder.get_connection().update.return_value = 1
        result = builder.from_('users').where('id', '=', 1).update(email='foo', name='bar')
        builder.get_connection().update.assert_called_with(
            query, ['foo', 'bar', 1]
        )
        self.assertEqual(1, result)

    def test_update_with_dictionaries(self):
        builder = self.get_builder()
        query = 'UPDATE "users" SET "email" = ?, "name" = ? WHERE "id" = ?'
        builder.get_connection().update.return_value = 1
        result = builder.from_('users').where('id', '=', 1).update({'email': 'foo', 'name': 'bar'})
        builder.get_connection().update.assert_called_with(
            query, ['foo', 'bar', 1]
        )
        self.assertEqual(1, result)
        builder = self.get_builder()

        query = 'UPDATE "users" SET "email" = ?, "name" = ? WHERE "id" = ?'
        builder.get_connection().update.return_value = 1
        result = builder.from_('users').where('id', '=', 1).update({'email': 'foo'}, name='bar')
        builder.get_connection().update.assert_called_with(
            query, ['foo', 'bar', 1]
        )
        self.assertEqual(1, result)

    def test_update_with_joins(self):
        builder = self.get_builder()
        query = 'UPDATE "users" ' \
                'INNER JOIN "orders" ON "users"."id" = "orders"."user_id" ' \
                'SET "email" = ?, "name" = ? WHERE "id" = ?'
        builder.get_connection().update.return_value = 1
        result = builder.from_('users')\
            .join('orders', 'users.id', '=', 'orders.user_id')\
            .where('id', '=', 1)\
            .update(email='foo', name='bar')
        builder.get_connection().update.assert_called_with(
            query, ['foo', 'bar', 1]
        )
        self.assertEqual(1, result)

    def test_update_on_postgres(self):
        builder = self.get_postgres_builder()
        marker = builder.get_grammar().get_marker()
        query = 'UPDATE "users" SET "email" = %s, "name" = %s WHERE "id" = %s' % (marker, marker, marker)
        builder.get_connection().update.return_value = 1
        result = builder.from_('users').where('id', '=', 1).update(email='foo', name='bar')
        builder.get_connection().update.assert_called_with(
            query, ['foo', 'bar', 1]
        )
        self.assertEqual(1, result)

    def test_update_with_joins_on_postgres(self):
        builder = self.get_postgres_builder()
        marker = builder.get_grammar().get_marker()
        query = 'UPDATE "users" ' \
                'SET "email" = %s, "name" = %s ' \
                'FROM "orders" WHERE "id" = %s AND "users"."id" = "orders"."user_id"'\
                % (marker, marker, marker)
        builder.get_connection().update.return_value = 1
        result = builder.from_('users')\
            .join('orders', 'users.id', '=', 'orders.user_id')\
            .where('id', '=', 1)\
            .update(email='foo', name='bar')
        builder.get_connection().update.assert_called_once_with(
            query, ['foo', 'bar', 1]
        )
        self.assertEqual(1, result)

    def test_update_respects_raw(self):
        builder = self.get_builder()
        marker = '?'
        query = 'UPDATE "users" SET "email" = foo, "name" = %s WHERE "id" = %s' % (marker, marker)
        builder.get_connection().update.return_value = 1
        result = builder.from_('users').where('id', '=', 1).update(email=QueryExpression('foo'), name='bar')
        builder.get_connection().update.assert_called_once_with(
            query, ['bar', 1]
        )
        self.assertEqual(1, result)

    def test_delete(self):
        builder = self.get_builder()
        query = 'DELETE FROM "users" WHERE "email" = ?'
        builder.get_connection().delete.return_value = 1
        result = builder.from_('users').where('email', '=', 'foo').delete()
        builder.get_connection().delete.assert_called_once_with(
            query, ['foo']
        )
        self.assertEqual(1, result)

        builder = self.get_builder()
        query = 'DELETE FROM "users" WHERE "id" = ?'
        builder.get_connection().delete.return_value = 1
        result = builder.from_('users').delete(1)
        builder.get_connection().delete.assert_called_once_with(
            query, [1]
        )
        self.assertEqual(1, result)

    def test_delete_with_join(self):
        builder = self.get_mysql_builder()
        marker = builder.get_grammar().get_marker()
        query = 'DELETE `users` FROM `users` ' \
                'INNER JOIN `contacts` ON `users`.`id` = `contacts`.`id` WHERE `email` = %s' % marker
        builder.get_connection().delete.return_value = 1
        result = builder.from_('users')\
            .join('contacts', 'users.id', '=', 'contacts.id')\
            .where('email', '=', 'foo')\
            .delete()
        builder.get_connection().delete.assert_called_once_with(
            query, ['foo']
        )
        self.assertEqual(1, result)

        builder = self.get_mysql_builder()
        marker = builder.get_grammar().get_marker()
        query = 'DELETE `users` FROM `users` ' \
                'INNER JOIN `contacts` ON `users`.`id` = `contacts`.`id` WHERE `id` = %s' % marker
        builder.get_connection().delete.return_value = 1
        result = builder.from_('users')\
            .join('contacts', 'users.id', '=', 'contacts.id')\
            .delete(1)
        builder.get_connection().delete.assert_called_once_with(
            query, [1]
        )
        self.assertEqual(1, result)

    def test_truncate(self):
        builder = self.get_builder()
        query = 'TRUNCATE "users"'
        builder.from_('users').truncate()
        builder.get_connection().statement.assert_called_once_with(
            query, []
        )

        builder = self.get_sqlite_builder()
        builder.from_('users')
        self.assertEqual({
            'DELETE FROM sqlite_sequence WHERE name = ?': ['users'],
            'DELETE FROM "users"': []
        }, builder.get_grammar().compile_truncate(builder))

    def test_postgres_insert_get_id(self):
        builder = self.get_postgres_builder()
        marker = builder.get_grammar().get_marker()
        query = 'INSERT INTO "users" ("email") VALUES (%s) RETURNING "id"' % marker
        builder.get_processor().process_insert_get_id.return_value = 1
        result = builder.from_('users').insert_get_id({'email': 'foo'}, 'id')
        builder.get_processor().process_insert_get_id.assert_called_once_with(
            builder, query, ['foo'], 'id'
        )
        self.assertEqual(1, result)

    def test_mysql_wrapping(self):
        builder = self.get_mysql_builder()
        builder.select('*').from_('users')
        self.assertEqual(
            'SELECT * FROM `users`',
            builder.to_sql()
        )

    def test_merge_wheres_can_merge_wheres_and_bindings(self):
        builder = self.get_builder()
        builder.wheres = ['foo']
        builder.merge_wheres(['wheres'], ['foo', 'bar'])
        self.assertEqual(['foo', 'wheres'], builder.wheres)
        self.assertEqual(['foo', 'bar'], builder.get_bindings())

    def test_where_with_null_second_parameter(self):
        builder = self.get_builder()
        builder.select('*').from_('users').where('foo', None)
        self.assertEqual(
            'SELECT * FROM "users" WHERE "foo" IS NULL',
            builder.to_sql()
        )

    def test_dynamic_where(self):
        method = 'where_foo_bar_and_baz_or_boom'
        parameters = ['john', 'jane', 'bam']

        builder = self.get_builder()
        builder.where = mock.MagicMock(return_value=builder)

        getattr(builder, method)(*parameters)

        builder.where.assert_has_calls([
            mock.call('foo_bar', '=', parameters[0], 'and'),
            mock.call('baz', '=', parameters[1], 'and'),
            mock.call('boom', '=', parameters[2], 'or')
        ])

    def test_dynamic_where_is_not_greedy(self):
        method = 'where_ios_version_and_android_version_or_orientation'
        parameters = ['6.1', '4.2', 'Vertical']

        builder = self.get_builder()
        builder.where = mock.MagicMock(return_value=builder)

        getattr(builder, method)(*parameters)

        builder.where.assert_has_calls([
            mock.call('ios_version', '=', parameters[0], 'and'),
            mock.call('android_version', '=', parameters[1], 'and'),
            mock.call('orientation', '=', parameters[2], 'or')
        ])

    def test_call_triggers_dynamic_where(self):
        builder = self.get_builder()

        self.assertEqual(builder, builder.where_foo_and_bar('baz', 'boom'))
        self.assertEqual(2, len(builder.wheres))

    def test_builder_raises_exception_with_undefined_method(self):
        builder = self.get_builder()

        try:
            builder.do_not_exist()
            self.fail('Builder did not raise and AttributeError exception')
        except AttributeError:
            self.assertTrue(True)

    def test_mysql_lock(self):
        builder = self.get_mysql_builder()
        marker = builder.get_grammar().get_marker()
        builder.select('*').from_('foo').where('bar', '=', 'baz').lock()
        self.assertEqual(
            'SELECT * FROM `foo` WHERE `bar` = %s FOR UPDATE' % marker,
            builder.to_sql()
        )
        self.assertEqual(['baz'], builder.get_bindings())

        builder = self.get_mysql_builder()
        marker = builder.get_grammar().get_marker()
        builder.select('*').from_('foo').where('bar', '=', 'baz').lock(False)
        self.assertEqual(
            'SELECT * FROM `foo` WHERE `bar` = %s LOCK IN SHARE MODE' % marker,
            builder.to_sql()
        )
        self.assertEqual(['baz'], builder.get_bindings())

    def test_postgres_lock(self):
        builder = self.get_postgres_builder()
        marker = builder.get_grammar().get_marker()
        builder.select('*').from_('foo').where('bar', '=', 'baz').lock()
        self.assertEqual(
            'SELECT * FROM "foo" WHERE "bar" = %s FOR UPDATE' % marker,
            builder.to_sql()
        )
        self.assertEqual(['baz'], builder.get_bindings())

        builder = self.get_postgres_builder()
        marker = builder.get_grammar().get_marker()
        builder.select('*').from_('foo').where('bar', '=', 'baz').lock(False)
        self.assertEqual(
            'SELECT * FROM "foo" WHERE "bar" = %s FOR SHARE' % marker,
            builder.to_sql()
        )
        self.assertEqual(['baz'], builder.get_bindings())

    def test_binding_order(self):
        expected_sql = 'SELECT * FROM "users" ' \
                       'INNER JOIN "othertable" ON "bar" = ? ' \
                       'WHERE "registered" = ? ' \
                       'GROUP BY "city" ' \
                       'HAVING "population" > ? ' \
                       'ORDER BY match ("foo") against(?)'
        expected_bindings = ['foo', True, 3, 'bar']

        builder = self.get_builder()
        builder.select('*').from_('users')\
            .order_by_raw('match ("foo") against(?)', ['bar'])\
            .where('registered', True)\
            .group_by('city')\
            .having('population', '>', 3)\
            .join(JoinClause('othertable').where('bar', '=', 'foo'))

        self.assertEqual(expected_sql, builder.to_sql())
        self.assertEqual(expected_bindings, builder.get_bindings())

    def test_add_binding_with_list_merges_bindings(self):
        builder = self.get_builder()
        builder.add_binding(['foo', 'bar'])
        builder.add_binding(['baz'])
        self.assertEqual(['foo', 'bar', 'baz'], builder.get_bindings())

    def test_add_binding_with_list_merges_bindings_in_correct_order(self):
        builder = self.get_builder()
        builder.add_binding(['bar', 'baz'], 'having')
        builder.add_binding(['foo'], 'where')
        self.assertEqual(['foo', 'bar', 'baz'], builder.get_bindings())

    def test_merge_builders(self):
        builder = self.get_builder()
        builder.add_binding('foo', 'where')
        builder.add_binding('baz', 'having')
        other_builder = self.get_builder()
        other_builder.add_binding('bar', 'where')
        builder.merge_bindings(other_builder)
        self.assertEqual(['foo', 'bar', 'baz'], builder.get_bindings())

    def test_sub_select(self):
        builder = self.get_builder()
        marker = builder.get_grammar().get_marker()
        expected_sql = 'SELECT "foo", "bar", (SELECT "baz" FROM "two" WHERE "subkey" = %s) AS "sub" ' \
                       'FROM "one" WHERE "key" = %s' % (marker, marker)
        expected_bindings = ['subval', 'val']

        builder.from_('one').select('foo', 'bar').where('key', '=', 'val')
        builder.select_sub(builder.new_query().from_('two').select('baz').where('subkey', '=', 'subval'), 'sub')
        self.assertEqual(expected_sql, builder.to_sql())
        self.assertEqual(expected_bindings, builder.get_bindings())

    def test_chunk(self):
        builder = self.get_builder()
        results = [
            {'foo': 'bar'},
            {'foo': 'baz'},
            {'foo': 'bam'},
            {'foo': 'boom'}
        ]

        def select(query, bindings, _):
            index = int(re.search('OFFSET (\d+)', query).group(1))
            limit = int(re.search('LIMIT (\d+)', query).group(1))

            if index >= len(results):
                return []

            return results[index:index + limit]

        builder.get_connection().select.side_effect = select

        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)

        i = 0
        for users in builder.from_('users').chunk(1):
            self.assertEqual(users[0], results[i])

            i += 1

        builder = self.get_builder()
        results = [
            {'foo': 'bar'},
            {'foo': 'baz'},
            {'foo': 'bam'},
            {'foo': 'boom'}
        ]

        builder.get_connection().select.side_effect = select

        builder.get_processor().process_select = mock.MagicMock(side_effect=lambda builder_, results_: results_)

        for users in builder.from_('users').chunk(2):
            self.assertEqual(2, len(users))

    def get_mysql_builder(self):
        grammar = MySqlQueryGrammar()
        processor = MockProcessor().prepare_mock()
        connection = MockConnection().prepare_mock()

        return QueryBuilder(connection, grammar, processor)

    def get_sqlite_builder(self):
        grammar = SQLiteQueryGrammar()
        processor = MockProcessor().prepare_mock()
        connection = MockConnection().prepare_mock()

        return QueryBuilder(connection, grammar, processor)

    def get_postgres_builder(self):
        grammar = PostgresQueryGrammar()
        processor = MockProcessor().prepare_mock()
        connection = MockConnection().prepare_mock()

        return QueryBuilder(connection, grammar, processor)

    def get_builder(self):
        grammar = QueryGrammar()
        processor = MockProcessor().prepare_mock()
        connection = MockConnection().prepare_mock()

        return QueryBuilder(connection, grammar, processor)



# -*- coding: utf-8 -*-

from .has_one_or_many import HasOneOrMany


class HasMany(HasOneOrMany):

    def get_results(self):
        """
        Get the results of the relationship.
        """
        return self._query.get()

    def init_relation(self, models, relation):
        """
        Initialize the relation on a set of models.

        :type models: list
        :type relation: str
        """
        for model in models:
            model.set_relation(relation, self._related.new_collection())

        return models

    def match(self, models, results, relation):
        """
        Match the eagerly loaded results to their parents.

        :type models: list
        :type results: Collection
        :type relation:  str
        """
        return self.match_many(models, results, relation)



# -*- coding: utf-8 -*-

from ..utils import PY2
from .connection import Connection
from ..query.grammars.postgres_grammar import PostgresQueryGrammar
from ..query.processors.postgres_processor import PostgresQueryProcessor


class PostgresConnection(Connection):

    def get_default_query_grammar(self):
        return PostgresQueryGrammar()

    def get_default_post_processor(self):
        return PostgresQueryProcessor()

    def begin_transaction(self):
        self._connection.autocommit = False

        super(PostgresConnection, self).begin_transaction()

    def commit(self):
        if self._transactions == 1:
            self._connection.commit()
            self._connection.autocommit = True

        self._transactions -= 1

    def rollback(self):
        if self._transactions == 1:
            self._transactions = 0

            self._connection.rollback()
            self._connection.autocommit = True
        else:
            self._transactions -= 1

    def _get_cursor_query(self, query, bindings):
        if not hasattr(self._cursor, 'query'):
            return super(PostgresConnection, self)._get_cursor_query(query, bindings)

        if PY2:
            return self._cursor.query

        return self._cursor.query.decode()



# -*- coding: utf-8 -*-

import threading
try:
    from Queue import Queue
except ImportError:
    from queue import Queue

from .. import EloquentTestCase
from .. import mock
from ..orm.models import User

from eloquent.query.builder import QueryBuilder
from eloquent.connections.connection import Connection


class ConnectionTestCase(EloquentTestCase):

    def test_table_returns_query_builder(self):
        connection = Connection(None, 'database')
        builder = connection.table('users')

        self.assertIsInstance(builder, QueryBuilder)
        self.assertEqual('users', builder.from__)
        self.assertEqual(connection.get_query_grammar(), builder.get_grammar())

    def test_transaction(self):
        connection = Connection(None, 'database')
        connection.begin_transaction = mock.MagicMock()
        connection.commit = mock.MagicMock()
        connection.rollback = mock.MagicMock()
        connection.insert = mock.MagicMock(return_value=1)

        with connection.transaction():
            connection.table('users').insert({'name': 'foo'})

        connection.begin_transaction.assert_called_once()
        connection.commit.assert_called_once()
        self.assertFalse(connection.rollback.called)

        connection.begin_transaction.reset_mock()
        connection.commit.reset_mock()
        connection.rollback.reset_mock()

        try:
            with connection.transaction():
                connection.table('users').insert({'name': 'foo'})
                raise Exception('foo')
        except Exception as e:
            self.assertEqual('foo', str(e))

        connection.begin_transaction.assert_called_once()
        connection.rollback.assert_called_once()
        self.assertFalse(connection.commit.called)


class ConnectionThreadLocalTest(EloquentTestCase):

    threads = 4

    def test_create_thread_local(self):
        self.init_database()

        def create_user_thread(low, hi):
            for _ in range(low, hi):
                User.create(name='u%d' % i)

            User.get_connection_resolver().disconnect()

        threads = []
        for i in range(self.threads):
            threads.append(threading.Thread(target=create_user_thread, args=(i*10, i * 10 + 10)))

        [t.start() for t in threads]
        [t.join() for t in threads]

        self.assertEqual(User.select().count(), self.threads * 10)

    def test_read_thread_local(self):
        self.init_database()

        data_queue = Queue()

        def reader_thread(q, num):
            for _ in range(num):
                data_queue.put(User.select().count())

        threads = []

        for i in range(self.threads):
            threads.append(threading.Thread(target=reader_thread, args=(data_queue, 20)))

        [t.start() for t in threads]
        [t.join() for t in threads]

        self.assertEqual(data_queue.qsize(), self.threads * 20)



# -*- coding: utf-8 -*-

from .relation import Relation
from .has_one import HasOne
from .has_many import HasMany
from .belongs_to import BelongsTo
from .belongs_to_many import BelongsToMany
from .has_many_through import HasManyThrough
from .morph_one import MorphOne
from .morph_many import MorphMany
from .morph_to import MorphTo
from .morph_to_many import MorphToMany



# -*- coding: utf-8 -*-


class Connector(object):

    RESERVED_KEYWORDS = [
        'log_queries', 'driver'
    ]

    def get_api(self):
        raise NotImplementedError()

    def get_config(self, config):
        return {x: config[x] for x in config if x not in self.RESERVED_KEYWORDS}

    def connect(self, config):
        return self.get_api().connect(**self.get_config(config))



# -*- coding: utf-8 -*-

import simplejson as json
import arrow
import inflection
import inspect
from six import add_metaclass
from ..exceptions.orm import MassAssignmentError
from ..query import QueryBuilder
from .builder import Builder
from .collection import Collection
from .relations import (
    Relation, HasOne, HasMany, BelongsTo, BelongsToMany, HasManyThrough,
    MorphOne, MorphMany, MorphTo, MorphToMany
)
from .relations.dynamic_property import DynamicProperty


class MetaModel(type):

    def __getattr__(cls, item):
        try:
            return type.__getattribute__(cls, item)
        except AttributeError:
            query = cls.query()

            return getattr(query, item)


@add_metaclass(MetaModel)
class Model(object):

    __connection__ = None

    __table__ = None

    __primary_key__ = 'id'

    __incrementing__ = True

    __fillable__ = []
    __guarded__ = ['*']
    __unguarded__ = False

    __hidden__ = []
    __visible__ = []

    __timestamps__ = True

    __casts__ = {}

    __touches__ = []

    __morph_class__ = None

    _with = []

    _booted = {}
    _registered = []

    __resolver = None

    many_methods = ['belongs_to_many', 'morph_to_many', 'morphed_by_many']

    CREATED_AT = 'created_at'
    UPDATED_AT = 'updated_at'

    def __init__(self, **attributes):
        """
        :param attributes: The instance attributes
        """
        self.__exists = False
        self.__dates = []
        self.__original = {}
        self.__attributes = {}
        self.__relations = {}

        self._boot_if_not_booted()

        self.sync_original()

        self.fill(**attributes)

    def _boot_if_not_booted(self):
        """
        Check if the model needs to be booted and if so, do it.
        """
        klass = self.__class__

        if not klass._booted.get(klass):
            klass._booted[klass] = True

            klass._boot()

    @classmethod
    def _boot(cls):
        """
        The booting method of the model.
        """
        # TODO

    def fill(self, **attributes):
        """
        Fill the model with attributes.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The model instance
        :rtype: Model

        :raises: MassAssignmentError
        """
        totally_guarded = self.totally_guarded()

        for key, value in self._fillable_from_dict(attributes).items():
            key = self._remove_table_from_key(key)

            if self.is_fillable(key):
                self.set_attribute(key, value)
            elif totally_guarded:
                raise MassAssignmentError(key)

        return self

    def force_fill(self, **attributes):
        """
        Fill the model with attributes. Force mass assignment.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The model instance
        :rtype: Model
        """
        self.unguard()

        self.fill(**attributes)

        self.reguard()

        return self

    def _fillable_from_dict(self, attributes):
        """
        Get the fillable attributes from a given dictionary.

        :type attributes: dict

        :return: The fillable attributes
        :rtype: dict
        """
        if self.__fillable__ and not self.__unguarded__:
            return {x: attributes[x] for x in attributes if x in self.__fillable__}

        return attributes

    def new_instance(self, attributes=None, exists=False):
        """
        Create a new instance for the given model.

        :param attributes: The instance attributes
        :type attributes: dict

        :param exists:
        :type exists: bool

        :return: A new instance for the current model
        :rtype: Model
        """
        if attributes is None:
            attributes = {}

        model = self.__class__(**attributes)

        model.set_exists(exists)

        return model

    def new_from_builder(self, attributes=None, connection=None):
        """
        Create a new model instance that is existing.

        :param attributes: The model attributes
        :type attributes: dict

        :param connection: The connection name
        :type connection: str

        :return: A new instance for the current model
        :rtype: Model
        """
        model = self.new_instance({}, True)

        if attributes is None:
            attributes = {}

        model.set_raw_attributes(attributes, True)

        model.set_connection(connection or self.__connection__)

        return model

    @classmethod
    def hydrate(cls, items, connection=None):
        """
        Create a collection of models from plain lists.

        :param items:
        :param connection:
        :return:
        """
        instance = cls().set_connection(connection)

        collection = instance.new_collection(items)

        return Collection(list(map(lambda item: instance.new_from_builder(item), collection)))

    @classmethod
    def hydrate_raw(cls, query, bindings=None, connection=None):
        """
        Create a collection of models from a raw query.

        :param query: The SQL query
        :type query: str

        :param bindings: The query bindings
        :type bindings: list

        :param connection: The connection name

        :rtype: Collection
        """
        instance = cls().set_connection(connection)

        items = instance.get_connection().select(query, bindings)

        return cls.hydrate(items, connection)

    @classmethod
    def create(cls, **attributes):
        """
        Save a new model an return the instance.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        model = cls(**attributes)

        model.save()

        return model

    @classmethod
    def force_create(cls, **attributes):
        """
        Save a new model an return the instance. Allow mass assignment.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        cls.unguard()

        model = cls.create(**attributes)

        cls.reguard()

        return model

    @classmethod
    def first_or_create(cls, **attributes):
        """
        Get the first record matching the attributes or create it.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        instance = cls.where(attributes).first()

        if instance is not None:
            return instance

        return cls.create(**attributes)

    @classmethod
    def first_or_new(cls, **attributes):
        """
        Get the first record matching the attributes or instantiate it.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        instance = cls.where(attributes).first()

        if instance is not None:
            return instance

        return cls(**attributes)

    @classmethod
    def update_or_create(cls, attributes, values=None):
        """
        Create or update a record matching the attributes, and fill it with values.

        :param attributes: The instance attributes
        :type attributes: dict

        :param values: The values
        :type values: dict

        :return: The new instance
        :rtype: Model
        """
        instance = cls.first_or_new(**attributes)

        if values is None:
            values = {}

        instance.fill(**values).save()

        return instance

    @classmethod
    def query(cls):
        """
        Begin querying the model.

        :return: A Builder instance
        :rtype: eloquent.orm.Builder
        """
        return cls().new_query()

    @classmethod
    def on(cls, connection=None):
        """
        Begin querying the model on a given connection.

        :param connection: The connection name
        :type connection: str

        :return: A Builder instance
        :rtype: eloquent.orm.Builder
        """
        instance = cls()

        instance.set_connection(connection)

        return instance.new_query()

    @classmethod
    def on_write_connection(cls):
        """
        Begin querying the model on the write connection.

        :return: A Builder instance
        :rtype: QueryBuilder
        """
        instance = cls()

        return instance.new_query().use_write_connection()

    @classmethod
    def all(cls, columns=None):
        """
        Get all og the models from the database.

        :param columns: The columns to retrieve
        :type columns: list

        :return: A Collection instance
        :rtype: Collection
        """
        instance = cls()

        return instance.new_query().get(columns)

    @classmethod
    def find(cls, id, columns=None):
        """
        Find a model by its primary key.

        :param id: The id of the model
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: Either a Model instance or a Collection
        :rtype: Model or Collection
        """
        instance = cls()

        if isinstance(id, list) and not id:
            return instance.new_collection()

        if columns is None:
            columns = ['*']

        return instance.new_query().find(id, columns)

    @classmethod
    def find_or_new(cls, id, columns=None):
        """
        Find a model by its primary key or return new instance.

        :param id: The id of the model
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: A Model instance
        :rtype: Model
        """
        instance = cls.find(id, columns)

        if instance is not None:
            return instance

        return cls()

    def fresh(self, with_=None):
        """
        Reload a fresh instance from the database.

        :param with_: The list of relations to eager load
        :type with_: list

        :return: The current model instance
        :rtype: Model
        """
        key = self.get_key_name()

        if self.exists:
            return self.__class__.with_(with_).where(key, self.get_key()).first()

    def load(self, relations):
        """
        Eager load relations on the model

        :param relations: The relations to eager load
        :type relations: str or list

        :return: The current model instance
        :rtype: Model
        """
        # TODO

    @classmethod
    def with_(cls, *relations):
        """
        Begin querying a model with eager loading

        :param relations: The relations to eager load
        :type relations: str or list

        :return: A Builder instance
        :rtype: Builder
        """
        instance = cls()

        return instance.new_query().with_(*relations)

    def has_one(self, related, foreign_key=None, local_key=None):
        """
        Define a one to one relationship.

        :param related: The related model:
        :type related: Model class

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param local_key: The local key
        :type local_key: str

        :rtype: HasOne
        """
        name = inspect.stack()[1][3]

        if name in self.__relations:
            return self.__relations[name]

        if not foreign_key:
            foreign_key = self.get_foreign_key()

        instance = related()

        if not local_key:
            local_key = self.get_key_name()

        return HasOne(instance.new_query(), self, '%s.%s' % (instance.get_table(), foreign_key), local_key)

    def morph_one(self, related, name, type_column=None, id_column=None, local_key=None):
        """
        Define a polymorphic one to one relationship.

        :param related: The related model:
        :type related: Model class

        :param type_column: The name of the type column
        :type type_column: str

        :param id_column: The name of the id column
        :type id_column: str

        :param local_key: The local key
        :type local_key: str

        :rtype: HasOne
        """
        name = inspect.stack()[1][3]

        if name in self.__relations:
            return self.__relations[name]

        instance = related()

        type_column, id_column = self.get_morphs(name, type_column, id_column)

        table = instance.get_table()

        if not local_key:
            local_key = self.get_key_name()

        return MorphOne(instance.new_query(), self,
                        '%s.%s' % (table, type_column),
                        '%s.%s' % (table, id_column), local_key)

    def belongs_to(self, related, foreign_key=None, other_key=None, relation=None):
        """
        Define an inverse one to one or many relationship.

        :param related: The related model:
        :type related: Model class

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :type relation: str

        :rtype: BelongsTo
        """
        if relation is None:
            relation = inspect.stack()[1][3]

        if relation in self.__relations:
            return self.__relations[relation]

        if foreign_key is None:
            foreign_key = '%s_id' % inflection.underscore(relation)

        instance = related()

        query = instance.new_query()

        if not other_key:
            other_key = instance.get_key_name()

        return BelongsTo(query, self, foreign_key, other_key, relation)

    def morph_to(self, name=None, type_column=None, id_column=None):
        """
        Define a polymorphic, inverse one-to-one or many relationship.

        :param name: The name of the relation
        :type name: str

        :param type_column: The type column
        :type type_column: str

        :param id_column: The id column
        :type id_column: str

        :rtype: MorphTo
        """
        if not name:
            name = inspect.stack()[1][3]

        if name in self.__relations:
            return self.__relations[name]

        type_column, id_column = self.get_morphs(name, type_column, id_column)

        if not hasattr(self, type_column):
            return MorphTo(self.new_query(), self, id_column, None, type_column, name)

        klass = None
        parent_type = getattr(self, type_column)
        for cls in Model.__subclasses__():
            morph_class = cls.__morph_class__ or cls.__name__
            if morph_class == parent_type:
                klass = cls
                break

        instance = klass()

        return MorphTo(instance.new_query(), self, id_column, instance.get_key_name(), type_column, name)

    def has_many(self, related, foreign_key=None, local_key=None):
        """
        Define a one to many relationship.

        :param related: The related model
        :type related: Model class

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param local_key: The local key
        :type local_key: str

        :rtype: HasOne
        """
        name = inspect.stack()[1][3]

        if name in self.__relations:
            return self.__relations[name]

        if not foreign_key:
            foreign_key = self.get_foreign_key()

        instance = related()

        if not local_key:
            local_key = self.get_key_name()

        return HasMany(instance.new_query(), self, '%s.%s' % (instance.get_table(), foreign_key), local_key)

    def has_many_through(self, related, through, first_key=None, second_key=None):
        """
        Define a has-many-through relationship.

        :param related: The related model
        :type related: Model class

        :param through: The through model
        :type through: Model class

        :param first_key: The first key
        :type first_key: str

        :param second_key: The second_key
        :type second_key: str

        :rtype: HasManyThrough
        """
        name = inspect.stack()[1][3]

        if name in self.__relations:
            return self.__relations[name]

        through = through()

        if not first_key:
            first_key = self.get_foreign_key()

        if not second_key:
            second_key = through.get_foreign_key()

        return HasManyThrough(related().new_query(), self, through, first_key, second_key)

    def morph_many(self, related, name, type_column=None, id_column=None, local_key=None):
        """
        Define a polymorphic one to many relationship.

        :param related: The related model:
        :type related: Model class

        :param type_column: The name of the type column
        :type type_column: str

        :param id_column: The name of the id column
        :type id_column: str

        :param local_key: The local key
        :type local_key: str

        :rtype: MorphMany
        """
        instance = related()

        if name in self.__relations:
            return self.__relations[name]

        type_column, id_column = self.get_morphs(name, type_column, id_column)

        table = instance.get_table()

        if not local_key:
            local_key = self.get_key_name()

        return MorphMany(instance.new_query(), self,
                         '%s.%s' % (table, type_column),
                         '%s.%s' % (table, id_column), local_key)

    def belongs_to_many(self, related, table=None, foreign_key=None, other_key=None, relation=None):
        """
        Define a many-to-many relationship.

        :param related: The related model:
        :type related: Model

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :type relation: str

        :rtype: BelongsToMany
        """
        if relation is None:
            relation = inspect.stack()[1][3]

        if relation in self.__relations:
            return self.__relations[relation]

        if not foreign_key:
            foreign_key = self.get_foreign_key()

        instance = related()

        if not other_key:
            other_key = instance.get_foreign_key()

        if table is None:
            table = self.joining_table(instance)

        query = instance.new_query()

        return BelongsToMany(query, self, table, foreign_key, other_key, relation)

    def morph_to_many(self, related, name, table=None, foreign_key=None, other_key=None, inverse=False):
        """
        Define a polymorphic many-to-many relationship.

        :param related: The related model:
        :type related: Model

        :param name: The relation name
        :type name: str

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :rtype: MorphToMany
        """
        caller = inspect.stack()[1][3]

        if caller in self.__relations:
            return self.__relations[caller]

        if not foreign_key:
            foreign_key = name + '_id'

        instance = related()

        if not other_key:
            other_key = instance.get_foreign_key()

        query = instance.new_query()

        if not table:
            table = inflection.pluralize(name)

        return MorphToMany(query, self, name, table,
                           foreign_key, other_key, caller, inverse)

    def morphed_by_many(self, related, name, table=None, foreign_key=None, other_key=None):
        """
        Define a polymorphic many-to-many relationship.

        :param related: The related model:
        :type related: Model

        :param name: The relation name
        :type name: str

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :rtype: MorphToMany
        """
        if not foreign_key:
            foreign_key = self.get_foreign_key()

        if not other_key:
            other_key = name + '_id'

        return self.morph_to_many(related, name, table, foreign_key, other_key, True)

    def joining_table(self, related):
        """
        Get the joining table name for a many-to-many relation

        :param related: The related model
        :type related: Model

        :rtype: str
        """
        base = self.get_table()

        related = related.get_table()

        models = sorted([related, base])

        return '_'.join(models)

    @classmethod
    def destroy(cls, *ids):
        """
        Destroy the models for the given IDs

        :param ids: The ids of the models to destroy
        :type ids: tuple

        :return: The number of models destroyed
        :rtype: int
        """
        count = 0

        if len(ids) == 1 and isinstance(ids[0], list):
            ids = ids[0]

        ids = list(ids)

        instance = cls()

        key = instance.get_key_name()

        for model in instance.new_query().where_in(key, ids).get():
            if model.delete():
                count += 1

        return count

    def delete(self):
        """
        Delete the model from the database.

        :rtype: bool or None

        :raises: Exception
        """
        if self.__primary_key__ is None:
            raise Exception('No primary key defined on the model.')

        if self.__exists:
            self._touch_owners()

            self._perform_delete_on_model()

            self.__exists = False

            return True

    def force_delete(self):
        """
        Force a hard delete on a soft deleted model.
        """
        return self.delete()

    def _perform_delete_on_model(self):
        """
        Perform the actual delete query on this model instance.
        """
        return self.new_query().where(self.get_key_name(), self.get_key()).delete()

    # TODO: events

    def _increment(self, column, amount=1):
        """
        Increment a column's value

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :return: The new column value
        :rtype: int
        """
        return self._increment_or_decrement(column, amount, 'increment')

    def _decrement(self, column, amount=1):
        """
        Decrement a column's value

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :return: The new column value
        :rtype: int
        """
        return self._increment_or_decrement(column, amount, 'decrement')

    def _increment_or_decrement(self, column, amount, method):
        """
        Runthe increment or decrement method on the model

        :param column: The column to increment or decrement
        :type column: str

        :param amount: The amount by which to increment or decrement
        :type amount: int

        :param method: The method
        :type method: str

        :return: The new column value
        :rtype: int
        """
        query = self.new_query()

        if not self.__exists:
            return getattr(query, method)(column, amount)

        self._increment_or_decrement_attribute_value(column, amount, method)

        query = query.where(self.get_key_name(), self.get_key())

        return getattr(query, method)(column, amount)

    def _increment_or_decrement_attribute_value(self, column, amount, method):
        """
        Increment the underlying attribute value and sync with original.

        :param column: The column to increment or decrement
        :type column: str

        :param amount: The amount by which to increment or decrement
        :type amount: int

        :param method: The method
        :type method: str

        :return: None
        """
        setattr(self, column, getattr(self, column) + (amount if method == 'increment' else amount * -1))

        self.sync_original_attribute(column)

    def update(self, **attributes):
        """
        Update the model in the database.

        :param attributes: The model attributes
        :type attributes: dict

        :return: The number of rows affected
        :rtype: int
        """
        if not self.__exists:
            return self.new_query().update(**attributes)

        return self.fill(**attributes).save()

    def push(self):
        """
        Save the model and all of its relationship.
        """
        if not self.save():
            return False

        for models in self.__relations.values():
            if isinstance(models, Collection):
                models = models.all()
            else:
                models = [models]

            for model in models:
                if not model:
                    continue

                if not model.push():
                    return False

        return True

    def save(self, options=None):
        """
        Save the model to the database.
        """
        if options is None:
            options = {}

        query = self.new_query()

        if self.__exists:
            saved = self._perform_update(query, options)
        else:
            saved = self._perform_insert(query, options)

        if saved:
            self._finish_save(options)

        return saved

    def _finish_save(self, options):
        """
        Finish processing on a successful save operation.
        """
        self.sync_original()

        if options.get('touch', True):
            self._touch_owners()

    def _perform_update(self, query, options=None):
        """
        Perform a model update operation.

        :param query: A Builder instance
        :type query: Builder

        :param options: Extra options
        :type options: dict
        """
        if options is None:
            options = {}

        dirty = self.get_dirty()

        if len(dirty):
            # TODO: "updating" event
            if self.__timestamps__ and options.get('timestamps', True):
                self._update_timestamps()

            dirty = self.get_dirty()

            if len(dirty):
                self._set_keys_for_save_query(query).update(dirty)

                # TODO: "updated" event

        return True

    def _perform_insert(self, query, options=None):
        """
        Perform a model update operation.

        :param query: A Builder instance
        :type query: Builder

        :param options: Extra options
        :type options: dict
        """
        if options is None:
            options = {}

        # TODO: "creating" event

        if self.__timestamps__ and options.get('timestamps', True):
            self._update_timestamps()

        attributes = self.__attributes

        if self.__incrementing__:
            self._insert_and_set_id(query, attributes)
        else:
            query.insert(attributes)

        self.__exists = True

        # TODO: "created" event

        return True

    def _insert_and_set_id(self, query, attributes):
        """
        Insert the given attributes and set the ID on the model.

        :param query: A Builder instance
        :type query: Builder

        :param attributes: The attributes to insert
        :type attributes: dict
        """
        key_name = self.get_key_name()

        id = query.insert_get_id(attributes, key_name)

        self.set_attribute(key_name, id)

    def _touch_owners(self):
        """
        Touch the owning relations of the model.
        """
        for relation in self.__touches__:
            if hasattr(self, relation):
                _relation = getattr(self, relation)
                _relation().touch()

                if _relation is not None:
                    _relation.touch_owners()

    def touches(self, relation):
        """
        Determine if a model touches a given relation.

        :param relation: The relation to check.
        :type relation: str

        :rtype: bool
        """
        return relation in self.__touches__

    def _set_keys_for_save_query(self, query):
        """
        Set the keys for a save update query.

        :param query: A Builder instance
        :type query: Builder

        :return: The Builder instance
        :rtype: Builder
        """
        query.where(self.get_key_name(), self._get_key_for_save_query())

        return query

    def _get_key_for_save_query(self):
        """
        Get the primary key value for a save query.
        """
        if self.get_key_name() in self.__original:
            return self.__original[self.get_key_name()]

        return self.__attributes[self.get_key_name()]

    def touch(self):
        """
        Update the model's timestamps.

        :rtype: bool
        """
        if not self.__timestamps__:
            return False

        self._update_timestamps()

        return self.save()

    def _update_timestamps(self):
        """
        Update the model's timestamps.
        """
        time = self.fresh_timestamp()

        if not self.is_dirty(self.UPDATED_AT):
            self.set_updated_at(time)

        if not self.__exists and not self.is_dirty(self.CREATED_AT):
            self.set_created_at(time)

    def set_created_at(self, value):
        """
        Set the value of the "created at" attribute.

        :param value: The value
        :type value: datetime
        """
        setattr(self, self.CREATED_AT, value)

    def set_updated_at(self, value):
        """
        Set the value of the "updated at" attribute.

        :param value: The value
        :type value: datetime
        """
        setattr(self, self.UPDATED_AT, value)

    def get_created_at_column(self):
        """
        Get the name of the "created at" column.

        :rtype: str
        """
        return self.CREATED_AT

    def get_updated_at_column(self):
        """
        Get the name of the "updated at" column.

        :rtype: str
        """
        return self.UPDATED_AT

    def fresh_timestamp(self):
        """
        Get a fresh timestamp for the model.

        :return: arrow.Arrow
        """
        return arrow.get().naive

    def new_query(self):
        """
        Get a new query builder for the model's table

        :return: A Builder instance
        :rtype: Builder
        """
        builder = self.new_orm_builder(
            self._new_base_query_builder()
        )

        return builder.set_model(self).with_(*self._with)

    @classmethod
    def query(cls):
        return cls().new_query()

    def new_orm_builder(self, query):
        """
        Create a new orm query builder for the model

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: A Builder instance
        :rtype: Builder
        """
        return Builder(query)

    def _new_base_query_builder(self):
        """
        Get a new query builder instance for the connection.

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        conn = self.get_connection()

        grammar = conn.get_query_grammar()

        return QueryBuilder(conn, grammar, conn.get_post_processor())

    def new_collection(self, models=None):
        """
        Create a new Collection instance.

        :param models: A list of models
        :type models: list

        :return: A new Collection instance
        :rtype: Collection
        """
        if models is None:
            models = []

        return Collection(models)

    def new_pivot(self, parent, attributes, table, exists):
        """
        Create a new pivot model instance.

        :param parent: The parent model
        :type parent: Model

        :param attributes: The pivot attributes
        :type attributes: dict

        :param table: the pivot table
        :type table: str

        :param exists: Whether the pivot exists or not
        :type exists: bool

        :rtype: Pivot
        """
        from .relations.pivot import Pivot

        return Pivot(parent, attributes, table, exists)

    def get_table(self):
        """
        Get the table associated with the model.

        :return: The name of the table
        :rtype: str
        """
        if self.__table__ is not None:
            return self.__table__

        return inflection.tableize(self.__class__.__name__)

    def set_table(self, table):
        """
        Set the table associated with the model.

        :param table: The table name
        :type table: str
        """
        self.__table__ = table

    def get_key(self):
        """
        Get the value of the model's primary key.
        """
        return self.get_attribute(self.get_key_name())

    def get_key_name(self):
        """
        Get the primary key for the model.

        :return: The primary key name
        :rtype: str
        """
        return self.__primary_key__

    def set_key_name(self, name):
        """
        Set the primary key for the model.

        :param name: The primary key name
        :type name: str
        """
        self.__primary_key__ = name

    def get_qualified_key_name(self):
        """
        Get the table qualified key name.

        :rtype: str
        """
        return '%s.%s' % (self.get_table(), self.get_key_name())

    def uses_timestamps(self):
        """
        Determine if the model uses timestamps.

        :rtype: bool
        """
        return self.__timestamps__

    def get_morphs(self, name, type, id):
        """
        Get the polymorphic relationship columns.
        """
        if not type:
            type = name + '_type'

        if not id:
            id = name + '_id'

        return type, id

    def get_morph_class(self):
        """
        Get the class name for polymorphic relations.
        """
        if not self.__morph_class__:
            return self.__class__.__name__

        return self.__morph_class__

    def get_foreign_key(self):
        """
        Get the default foreign key name for the model

        :rtype: str
        """
        return '%s_id' % inflection.singularize(inflection.tableize(self.__class__.__name__))

    def get_hidden(self):
        """
        Get the hidden attributes for the model.
        """
        return self.__hidden__

    def set_hidden(self, hidden):
        """
        Set the hidden attributes for the model.

        :param hidden: The attributes to add
        :type hidden: list
        """
        self.__hidden__ = hidden

        return self

    def add_hidden(self, *attributes):
        """
        Add hidden attributes to the model.

        :param attributes: The attributes to hide
        :type attributes: list
        """
        self.__hidden__ += attributes

    def get_visible(self):
        """
        Get the visible attributes for the model.
        """
        return self.__visible__

    def set_visible(self, visible):
        """
        Set the visible attributes for the model.

        :param visible: The attributes to make visible
        :type visible: list
        """
        self.__visible__ = visible

        return self

    def add_visible(self, *attributes):
        """
        Add visible attributes to the model.

        :param attributes: The attributes to make visible
        :type attributes: list
        """
        self.__visible__ += attributes

    def get_fillable(self):
        """
        Get the fillable attributes for the model.

        :rtype: list
        """
        return self.__fillable__

    def fillable(self, fillable):
        """
        Set the fillable attributes for the model.

        :param fillable: The fillable attributes
        :type fillable: list

        :return: The current Model instance
        :rtype: Model
        """
        self.__fillable__ = fillable

        return self

    def get_guarded(self):
        """
        Get the guarded attributes.
        """
        return self.__guarded__

    def guard(self, guarded):
        """
        Set the guarded attributes.

        :param guarded: The guarded attributes
        :type guarded: list

        :return: The current Model instance
        :rtype: Model
        """
        self.__guarded__ = guarded

        return self

    @classmethod
    def unguard(cls):
        """
        Disable the mass assigment restrictions.
        """
        cls.__unguarded__ = True

    @classmethod
    def reguard(cls):
        """
        Enable the mass assignment restrictions.
        :return:
        """
        cls.__unguarded__ = False

    def is_fillable(self, key):
        """
        Determine if the given attribute can be mass assigned.

        :param key: The attribute to check
        :type key: str

        :return: Whether the attribute can be mass assigned or not
        :rtype: bool
        """
        if self.__unguarded__:
            return True

        if key in self.__fillable__:
            return True

        if self.is_guarded(key):
            return False

        return not self.__fillable__ and not key.startswith('_')

    def is_guarded(self, key):
        """
        Determine if the given attribute is guarded.

        :param key: The attribute to check
        :type key: str

        :return: Whether the attribute is guarded or not
        :rtype: bool
        """
        return key in self.__guarded__ or self.__guarded__ == ['*']

    def totally_guarded(self):
        """
        Determine if the model is totally guarded.

        :rtype: bool
        """
        return len(self.__fillable__) == 0 and self.__guarded__ == ['*']

    def _remove_table_from_key(self, key):
        """
        Remove the table name from a given key.

        :param key: The key to remove the table name from.
        :type key: str

        :rtype: str
        """
        if '.' not in key:
            return key

        return key.split('.')[-1]

    def get_incrementing(self):
        return self.__incrementing__

    def set_incrementing(self, value):
        self.__incrementing__ = value

    def to_json(self, **options):
        """
        Convert the model instance to JSON.

        :param options: The JSON options
        :type options: dict

        :return: The JSON encoded model instance
        :rtype: str
        """
        return json.dumps(self.to_dict(), **options)

    def json_serialize(self):
        """
        Convert the object into something JSON serializable.

        :rtype: dict
        """
        return self.to_dict()

    def to_dict(self):
        """
        Convert the model instance to a dictionary.

        :return: The dictionary version of the model instance
        :rtype: dict
        """
        attributes = self.attributes_to_dict()

        attributes.update(self.relations_to_dict())

        return attributes

    def attributes_to_dict(self):
        """
        Convert the model's attributes to a dictionary.

        :rtype: dict
        """
        attributes = self._get_dictable_attributes()

        for key in self.get_dates():
            if not key in attributes:
                continue

            attributes[key] = self._format_date(self.as_datetime(attributes[key]))

        # TODO: mutators

        for key, value in self.__casts__.items():
            if key not in attributes:  # TODO: check mutators
                continue

            attributes[key] = self._cast_attribute(key, attributes[key])

        # TODO: appends

        return attributes

    def _get_dictable_attributes(self):
        """
        Get an attribute dictionary of all dictable attributes.

        :rtype: dict
        """
        return self._get_dictable_items(self.__attributes)

    def relations_to_dict(self):
        """
        Get the model's relationships in dictionary form.

        :rtype: dict
        """
        attributes = {}

        for key, value in self._get_dictable_relations().items():
            if key in self.get_hidden():
                continue

            relation = None
            if hasattr(value, 'to_dict'):
                relation = value.to_dict()
            elif value is None:
                relation = value

            if relation or value is None:
                attributes[key] = relation

        return attributes

    def _get_dictable_relations(self):
        """
        Get an attribute dict of all dictable relations.
        """
        return self._get_dictable_items(self.__relations)

    def _get_dictable_items(self, values):
        """
        Get an attribute dictionary of all dictable values.

        :param values: The values to check
        :type values: dict

        :rtype: dict
        """
        if len(self.get_visible()) > 0:
            return {x: values[x] for x in values.keys() if x in self.get_visible()}

        return {x: values[x] for x in values.keys() if x not in self.get_hidden() and not x.startswith('_')}

    def get_attribute(self, key, original=None):
        """
        Get an attribute from the model.

        :param key: The attribute to get
        :type key: str
        """
        in_attributes = key in self.__attributes

        if in_attributes:
            return self._get_attribute_value(key)

        if key in self.__relations:
            return self.__relations[key]

        relation = original or super(Model, self).__getattribute__(key)

        if relation:
            return self._get_relationship_from_method(key, relation)

        raise AttributeError(key)

    def _get_attribute_value(self, key):
        """
        Get a plain attribute.

        :param key: The attribute to get
        :type key: str
        """
        value = self._get_attribute_from_dict(key)

        # TODO: mutators

        if self._has_cast(key):
            value = self._cast_attribute(key, value)
        elif key in self.get_dates():
            if value is not None:
                return self.as_datetime(value)

        return value

    def _get_attribute_from_dict(self, key):
        return self.__attributes.get(key)

    def _get_relationship_from_method(self, method, relations=None):
        """
        Get a relationship value from a method.

        :param method: The method name
        :type method: str

        :rtype: mixed
        """
        relations = relations or super(Model, self).__getattribute__(method)

        if not isinstance(relations, Relation):
            raise RuntimeError('Relationship method must return an object of type Relation')

        self.__relations[method] = DynamicProperty(relations.get_results, relations)

        return self.__relations[method]

    def has_get_mutator(self, key):
        """
        Determine if a get mutator exists for an attribute.

        :param key: The attribute name
        :type key: str

        :rtype: bool
        """
        return hasattr(self, 'get_%s_attribute' % inflection.underscore(key))

    def _has_cast(self, key):
        """
        Determine whether an attribute should be casted to a native type.

        :param key: The attribute to check
        :type key: str

        :rtype: bool
        """
        return key in self.__casts__

    def _is_json_castable(self, key):
        """
        Determine whether a value is JSON castable.

        :param key: The key to check
        :type key: str

        :rtype: bool
        """
        if self._has_cast(key):
            type = self._get_cast_type(key)

            return type in ['list', 'dict', 'json', 'object']

        return False

    def _get_cast_type(self, key):
        """
        Get the type of the cast for a model attribute.

        :param key: The attribute to get the cast for
        :type key: str

        :rtype: str
        """
        return self.__casts__[key].lower().strip()

    def _cast_attribute(self, key, value):
        """
        Cast an attribute to a native Python type

        :param key: The attribute key
        :type key: str

        :param value: The attribute value
        :type value: The attribute value

        :rtype: mixed
        """
        if value is None:
            return None

        type = self._get_cast_type(key)
        if type in ['int', 'integer']:
            return int(value)
        elif type in ['real', 'float', 'double']:
            return float(value)
        elif type in ['string', 'str']:
            return str(value)
        elif type in ['bool', 'boolean']:
            return bool(value)
        elif type in ['dict', 'list', 'json']:
            return json.loads(value)
        else:
            return value

    def get_dates(self):
        """
        Get the attributes that should be converted to dates.

        :rtype: list
        """
        defaults = [self.CREATED_AT, self.UPDATED_AT]

        return self.__dates + defaults

    def from_datetime(self, value):
        """
        Convert datetime to a datetime object

        :rtype: datetime.datetime
        """
        if isinstance(value, arrow.Arrow):
            return value.naive

        return arrow.get(value).naive

    def as_datetime(self, value):
        """
        Return a timestamp as a datetime.

        :rtype: arrow.Arrow
        """
        return arrow.get(value)

    def get_date_format(self):
        """
        Get the format to use for timestamps and dates.

        :rtype: str
        """
        return 'iso'

    def _format_date(self, date):
        """
        Format a date or timestamp.

        :param date: The date or timestamp
        :type date: datetime.datetime or datetime.date or arrow.Arrow

        :rtype: str
        """
        format = self.get_date_format()

        if format == 'iso':
            return date.isoformat()
        else:
            if isinstance(date, arrow.Arrow):
                return date.format(format)

            return date.strftime(format)

    def set_attribute(self, key, value):
        """
        Set a given attribute on the model.
        """
        # TODO: Set mutators

        if key in self.get_dates() and value:
            value = self.from_datetime(value)

        if self._is_json_castable(key):
            value = json.dumps(value)

        self.__attributes[key] = value

    def replicate(self, except_=None):
        """
        Clone the model into a new, non-existing instance.

        :param except_: The attributes that should not be cloned
        :type except_: list

        :rtype: Model
        """
        if except_ is None:
            except_ = [
                self.get_key_name(),
                self.get_created_at_column(),
                self.get_updated_at_column()
            ]

            attributes = {x: self.__attributes[x] for x in self.__attributes if x not in except_}

            instance = self.new_instance(attributes)

            instance.set_relations(dict(**self.__relations))

            return instance

    def get_attributes(self):
        """
        Get all of the current attributes on the model.

        :rtype: dict
        """
        return self.__attributes

    def set_raw_attributes(self, attributes, sync=False):
        """
        Set the dictionary of model attributes. No checking is done.

        :param attributes: The model attributes
        :type attributes: dict

        :param sync: Whether to sync the attributes or not
        :type sync: bool
        """
        self.__attributes = dict(attributes.items())

        if sync:
            self.sync_original()

    def get_original(self, key=None, default=None):
        """
        Get the original values

        :param key: The original key to get
        :type key: str

        :param default: The default value if the key does not exist
        :type default: mixed

        :rtype: mixed
        """
        if key is None:
            return self.__original

        return self.__original.get(key, default)

    def sync_original(self):
        """
        Sync the original attributes with the current.

        :rtype: Builder
        """
        self.__original = dict(self.__attributes.items())

        return self

    def sync_original_attribute(self, attribute):
        """
        Sync a single original attribute with its current value.

        :param attribute: The attribute to sync
        :type attribute: str

        :rtype: Model
        """
        self.__original[attribute] = self.__attributes[attribute]

        return self

    def is_dirty(self, *attributes):
        """
        Determine if the model or given attributes have been modified.

        :param attributes: The attributes to check
        :type attributes: list

        :rtype: boolean
        """
        dirty = self.get_dirty()

        if not attributes:
            return len(dirty) > 0

        for attribute in attributes:
            if attribute in dirty:
                return True

        return False

    def get_dirty(self):
        """
        Get the attribute that have been change since last sync.

        :rtype: list
        """
        dirty = {}

        for key, value in self.__attributes.items():
            if key not in self.__original:
                dirty[key] = value
            elif value != self.__original[key]:
                dirty[key] = value

        return dirty

    @property
    def exists(self):
        return self.__exists

    def set_exists(self, exists):
        self.__exists = exists

    def get_relations(self):
        """
        Get all the loaded relations for the instance.

        :rtype: dict
        """
        return self.__relations

    def get_relation(self, relation):
        """
        Get a specific relation.

        :param relation: The name of the relation.
        :type relation: str

        :rtype: mixed
        """
        return self.__relations[relation]

    def set_relation(self, relation, value):
        """
        Set the specific relation in the model.

        :param relation: The name of the relation
        :type relation: str

        :param value: The relation
        :type value: mixed

        :return: The current Model instance
        :rtype: Model
        """
        self.__relations[relation] = value

        return self

    def set_relations(self, relations):
        self.__relations = relations

        return self

    def get_connection(self):
        """
        Get the database connection for the model

        :rtype: eloquent.connections.Connection
        """
        return self.resolve_connection(self.__connection__)

    def get_connection_name(self):
        """
        Get the database connection name for the model.

        :rtype: str
        """
        return self.__connection__

    def set_connection(self, name):
        """
        Set the connection associated with the model.

        :param name: The connection name
        :type name: str

        :return: The current model instance
        :rtype: Model
        """
        self.__connection__ = name

        return self

    @classmethod
    def resolve_connection(cls, connection=None):
        """
        Resolve a connection instance.

        :param connection: The connection name
        :type connection: str

        :rtype: eloquent.connections.Connection
        """
        return cls.__resolver.connection(connection)

    @classmethod
    def get_connection_resolver(cls):
        """
        Get the connection resolver instance.
        """
        return cls.__resolver

    @classmethod
    def set_connection_resolver(cls, resolver):
        """
        Set the connection resolver instance.
        """
        cls.__resolver = resolver

    @classmethod
    def unset_connection_resolver(cls, resolver):
        """
        Unset the connection resolver instance.
        """
        cls._resolver = None

    def __getattribute__(self, item):
        try:
            attr = super(Model, self).__getattribute__(item)
            if isinstance(attr, Relation):
                return self.get_attribute(item, attr)

            return attr
        except AttributeError:
            return self.get_attribute(item)

    def __setattr__(self, key, value):
        if key.startswith(('_Model__', '_%s__' % self.__class__.__name__, '__')):
            super(Model, self).__setattr__(key, value)
        elif callable(getattr(self, key, None)):
            return super(Model, self).__setattr__(key, value)
        else:
            self.set_attribute(key, value)

    def __delattr__(self, item):
        try:
            super(Model, self).__delattr__(item)
        except AttributeError:
            del self.__attributes[item]



# -*- coding: utf-8 -*-

from .base import BaseQueryGrammar
from .grammar import QueryGrammar
from .postgres_grammar import PostgresQueryGrammar
from .mysql_grammar import MySqlQueryGrammar
from .sqlite_grammar import SQLiteQueryGrammar



# -*- coding: utf-8 -*-

from .connection import Connection
from ..query.grammars.mysql_grammar import MySqlQueryGrammar
from ..query.processors.mysql_processor import MySqlQueryProcessor


class MySqlConnection(Connection):

    def get_default_query_grammar(self):
        return MySqlQueryGrammar()

    def get_default_post_processor(self):
        return MySqlQueryProcessor()

    def begin_transaction(self):
        self._connection.autocommit(False)

        super(MySqlConnection, self).begin_transaction()

    def commit(self):
        if self._transactions == 1:
            self._connection.commit()
            self._connection.autocommit(True)

        self._transactions -= 1

    def rollback(self):
        if self._transactions == 1:
            self._transactions = 0

            self._connection.rollback()
            self._connection.autocommit(True)
        else:
            self._transactions -= 1

    def _get_cursor_query(self, query, bindings):
        if not hasattr(self._cursor, '_last_executed'):
            return super(MySqlConnection, self)._get_cursor_query(query, bindings)

        return self._cursor._last_executed



# -*- coding: utf-8 -*-

from ..collection import Collection
from .relation import Relation


class HasOneOrMany(Relation):

    def __init__(self, query, parent, foreign_key, local_key):
        """
        :type query: eloquent.orm.Builder

        :param parent: The parent model
        :type parent: Model

        :param foreign_key: The foreign key of the parent model
        :type foreign_key: str

        :param local_key: The local key of the parent model
        :type local_key: str
        """
        self._local_key = local_key
        self._foreign_key = foreign_key

        super(HasOneOrMany, self).__init__(query, parent)

    def add_constraints(self):
        """
        Set the base constraints of the relation query
        """
        if self._constraints:
            self._query.where(self._foreign_key, '=', self.get_parent_key())

    def add_eager_constraints(self, models):
        """
        Set the constraints for an eager load of the relation.

        :type models: list
        """
        return self._query.where_in(self._foreign_key, self.get_keys(models, self._local_key))

    def match_one(self, models, results, relation):
        """
        Match the eargerly loaded resuls to their single parents.

        :param models: The parents
        :type models: list

        :param results: The results collection
        :type results: Collection

        :param relation: The relation
        :type relation: str

        :rtype: list
        """
        return self._match_one_or_many(models, results, relation, 'one')

    def match_many(self, models, results, relation):
        """
        Match the eargerly loaded resuls to their single parents.

        :param models: The parents
        :type models: list

        :param results: The results collection
        :type results: Collection

        :param relation: The relation
        :type relation: str

        :rtype: list
        """
        return self._match_one_or_many(models, results, relation, 'many')

    def _match_one_or_many(self, models, results, relation, type):
        """
        Match the eargerly loaded resuls to their single parents.

        :param models: The parents
        :type models: list

        :param results: The results collection
        :type results: Collection

        :param relation: The relation
        :type relation: str

        :param type: The match type
        :type type: str

        :rtype: list
        """
        dictionary = self._build_dictionary(results)

        for model in models:
            key = model.get_attribute(self._local_key)

            if key in dictionary:
                value = self._get_relation_value(dictionary, key, type)

                model.set_relation(relation, value)

        return models

    def _get_relation_value(self, dictionary, key, type):
        """
        Get the value of the relationship by one or many type.

        :type dictionary: dict
        :type key: str
        :type type: str
        """
        value = dictionary[key]

        if type == 'one':
            return value[0]

        return self._related.new_collection(value)

    def _build_dictionary(self, results):
        """
        Build model dictionary keyed by the relation's foreign key.

        :param results: The results
        :type results: Collection

        :rtype: dict
        """
        dictionary = {}

        foreign = self.get_plain_foreign_key()

        for result in results:
            key = getattr(result, foreign)
            if key not in dictionary:
                dictionary[key] = []

            dictionary[key].append(result)

        return dictionary

    def save(self, model):
        """
        Attach a model instance to the parent models.

        :param model: The model instance to attach
        :type model: Model

        :rtype: Model
        """
        model.set_attribute(self.get_plain_foreign_key(), self.get_parent_key())

        if model.save():
            return model

        return False

    def save_many(self, models):
        """
        Attach a list of models to the parent instance.

        :param models: The models to attach
        :type models: list of Model

        :rtype: list
        """
        return map(self.save, models)

    def find_or_new(self, id, columns=None):
        """
        Find a model by its primary key or return new instance of the related model.

        :param id: The primary key
        :type id: mixed

        :param columns:  The columns to retrieve
        :type columns: list

        :rtype: Collection or Model
        """
        if columns is None:
            columns = ['*']

        instance = self.find(id, columns)

        if instance is None:
            instance = self._related.new_instance()
            instance.set_attribute(self.get_plain_foreign_key(), self.get_parent_key())

        return instance

    def first_or_new(self, _attributes=None, **attributes):
        """
        Get the first related model record matching the attributes or instantiate it.

        :param attributes:  The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self.where(attributes).first()

        if instance is None:
            instance = self._related.new_instance()
            instance.set_attribute(self.get_plain_foreign_key(), self.get_parent_key())

        return instance

    def first_or_create(self, _attributes=None, **attributes):
        """
        Get the first related record matching the attributes or create it.

        :param attributes:  The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self.where(attributes).first()

        if instance is None:
            instance = self.create(**attributes)

        return instance

    def update_or_create(self, attributes, values=None):
        """
        Create or update a related record matching the attributes, and fill it with values.

        :param attributes: The attributes
        :type attributes: dict

        :param values: The values
        :type values: dict

        :rtype: Model
        """
        instance = self.first_or_new(**attributes)

        instance.fill(values)

        instance.save()

        return instance

    def create(self, _attributes=None, **attributes):
        """
        Create a new instance of the related model.

        :param attributes: The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self._related.new_instance(attributes)

        instance.set_attribute(self.get_plain_foreign_key(), self.get_parent_key())

        instance.save()

        return instance

    def create_many(self, records):
        """
        Create a list of new instances of the related model.

        :param records: instances attributes
        :type records: list

        :rtype: list
        """
        instances = []

        for record in records:
            instances.append(self.create(**record))

        return instances

    def update(self, _attributes=None, **attributes):
        """
        Perform an update on all the related models.

        :param attributes: The attributes
        :type attributes: dict

        :rtype: int
        """
        if _attributes is not None:
            attributes.update(_attributes)

        if self._related.uses_timestamps():
            attributes[self.get_related_updated_at()] = self._related.fresh_timestamp()

        return self._query.update(attributes)

    def get_has_compare_key(self):
        return self.get_foreign_key()

    def get_foreign_key(self):
        return self._foreign_key

    def get_plain_foreign_key(self):
        segments = self.get_foreign_key().split('.')

        return segments[-1]

    def get_parent_key(self):
        return self._parent.get_attribute(self._local_key)

    def get_qualified_parent_key_name(self):
        return '%s.%s' % (self._parent.get_table(), self._local_key)



# -*- coding: utf-8 -*-

from . import EloquentTestCase
from . import mock
from .utils import MockConnection, MockManager, MockFactory

from eloquent.database_manager import DatabaseManager


class ConnectionTestCase(EloquentTestCase):

    def test_connection_method_create_a_new_connection_if_needed(self):
        manager = self._get_manager()
        manager.table('users')

        manager._make_connection.assert_called_once_with(
            'sqlite'
        )

        manager._make_connection.reset_mock()
        manager.table('users')
        self.assertFalse(manager._make_connection.called)

    def test_manager_uses_factory_to_create_connections(self):
        manager = self._get_real_manager()
        manager.connection()

        manager._factory.make.assert_called_with(
            {
                'driver': 'sqlite',
                'database': ':memory:'
            }, 'sqlite'
        )

    def test_connection_can_select_connections(self):
        manager = self._get_manager()
        self.assertEqual(manager.connection(), manager.connection('sqlite'))
        self.assertNotEqual(manager.connection('sqlite'), manager.connection('sqlite2'))

    def test_dynamic_attribute_gets_connection_attribute(self):
        manager = self._get_manager()
        manager.statement('CREATE TABLE users')

        manager.get_connections()['sqlite'].statement.assert_called_once_with(
            'CREATE TABLE users'
        )

    def test_default_database_with_one_database(self):
        manager = MockManager({
            'sqlite': {
                'driver': 'sqlite',
                'database': ':memory:'
            }
        }).prepare_mock()

        self.assertEqual('sqlite', manager.get_default_connection())

    def _get_manager(self):
        manager = MockManager({
            'default': 'sqlite',
            'sqlite': {
                'driver': 'sqlite',
                'database': ':memory:'
            },
            'sqlite2': {
                'driver': 'sqlite',
                'database': ':memory:'
            }
        }).prepare_mock()

        return manager

    def _get_real_manager(self):
        manager = DatabaseManager({
            'default': 'sqlite',
            'sqlite': {
                'driver': 'sqlite',
                'database': ':memory:'
            }
        }, MockFactory().prepare_mock())

        return manager

    def _get_connection(self):
        return MockConnection().prepare_mock()



# -*- coding: utf-8 -*-

from .grammar import QueryGrammar


class SQLiteQueryGrammar(QueryGrammar):

    _operators = [
        '=', '<', '>', '<=', '>=', '<>', '!=',
        'like', 'not like', 'between', 'ilike',
        '&', '|', '<<', '>>',
    ]

    def compile_insert(self, query, values):
        """
        Compile insert statement into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param values: The insert values
        :type values: dict or list

        :return: The compiled insert
        :rtype: str
        """
        table = self.wrap_table(query.from__)

        if not isinstance(values, list):
            values = [values]

        # If there is only one row to insert, we just use the normal grammar
        if len(values) == 1:
            return super(SQLiteQueryGrammar, self).compile_insert(query, values)

        names = self.columnize(values[0].keys())

        columns = []

        # SQLite requires us to build the multi-row insert as a listing of select with
        # unions joining them together. So we'll build out this list of columns and
        # then join them all together with select unions to complete the queries.
        for column in values[0].keys():
            columns.append('%s AS %s' % (self.get_marker(), self.wrap(column)))

        columns = [', '.join(columns)] * len(values)

        return 'INSERT INTO %s (%s) SELECT %s'\
               % (table, names, ' UNION ALL SELECT '.join(columns))

    def compile_truncate(self, query):
        """
        Compile a truncate statement into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled truncate statement
        :rtype: str
        """
        sql = {
            'DELETE FROM sqlite_sequence WHERE name = %s' % self.get_marker(): [query.from__]
        }

        sql['DELETE FROM %s' % self.wrap_table(query.from__)] = []

        return sql

    def _where_day(self, query, where):
        """
        Compile a "where day" clause

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param where: The condition
        :type where: dict

        :return: The compiled clause
        :rtype: str
        """
        return self._date_based_where('%d', query, where)

    def _where_month(self, query, where):
        """
        Compile a "where month" clause

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param where: The condition
        :type where: dict

        :return: The compiled clause
        :rtype: str
        """
        return self._date_based_where('%m', query, where)

    def _where_year(self, query, where):
        """
        Compile a "where year" clause

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param where: The condition
        :type where: dict

        :return: The compiled clause
        :rtype: str
        """
        return self._date_based_where('%Y', query, where)

    def _date_based_where(self, type, query, where):
        """
        Compiled a date where based clause

        :param type: The date type
        :type type: str

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param where: The condition
        :type where: dict

        :return: The compiled clause
        :rtype: str
        """
        value = str(where['value']).zfill(2)
        value = self.parameter(value)

        return 'strftime(\'%s\', %s) %s %s'\
               % (type, self.wrap(where['column']),
                  where['operator'], value)

    def get_marker(self):
        return '?'



# -*- coding: utf-8 -*-


class ConnectorException(Exception):

    pass


class UnsupportedDriver(ConnectorException):

    def __init__(self, driver):
        message = 'Driver "%s" is not supported' % driver

        super(UnsupportedDriver, self).__init__(message)



# -*- coding: utf-8 -*-


class DynamicProperty(object):
    """
    Relationship dynamic property.

    It provides a simple way to access a property as is, returning the results,
    or has a method whihch will start a query on the relation.

    Example:

    >>> user = User.find(1)
    >>> user.roles  # will return the roles associated with the user
    >>> user.roles().first() # Will return the first role
    """

    def __init__(self, results_getter, relation):
        self._results_getter = results_getter
        self._results = None
        self._relation = relation

    def get_results(self):
        return self._results

    def __getitem__(self, item):
        if not self._results:
            self._results = self._results_getter()

        return self._results[item]

    def __iter__(self):
        if not self._results:
            self._results = self._results_getter()

        return iter(self._results)

    def __getattr__(self, item):
        if not self._results:
            self._results = self._results_getter()

        return getattr(self._results, item)

    def __call__(self, *args, **kwargs):
        return self._relation(*args, **kwargs)



# -*- coding: utf-8 -*-

import random
from ..exceptions import ArgumentError
from ..exceptions.connectors import UnsupportedDriver
from .mysql_connector import MySqlConnector
from .postgres_connector import PostgresConnector
from .sqlite_connector import SQLiteConnector
from ..connections import (
    MySqlConnection,
    PostgresConnection,
    SQLiteConnection
)


class ConnectionFactory(object):

    def make(self, config, name=None):
        if 'read' in config:
            return self._create_read_write_connection(config)

        return self._create_single_connection(config)

    def _create_single_connection(self, config):
        conn = self.create_connector(config).connect(config)

        return self._create_connection(
            config['driver'],
            conn,
            config['database'],
            config.get('prefix'),
            config
        )

    def _create_read_write_connection(self, config):
        connection = self._create_single_connection(config)

        connection.set_read_connection(self._create_read_connection(config))

        return connection

    def _create_read_connection(self, config):
        read_config = self._get_read_config(config)

        return self.create_connector(read_config).connect(read_config)

    def _get_read_config(self, config):
        read_config = self._get_read_write_config(config, 'read')

        return self._merge_read_write_config(config, read_config)

    def _get_write_config(self, config):
        write_config = self._get_read_write_config(config, 'write')

        return self._merge_read_write_config(config, write_config)

    def _get_read_write_config(self, config, type):
        if config.get(type, []):
            return random.choice(config[type])

        return config[type]

    def _merge_read_write_config(self, config, merge):
        config = config.copy()
        config.update(merge)

        del config['read']
        del config['write']

        return config

    def create_connector(self, config):
        if 'driver' not in config:
            raise ArgumentError('A driver must be specified')

        driver = config['driver']

        if driver == 'mysql':
            return MySqlConnector()
        elif driver == 'postgres':
            return PostgresConnector()
        elif driver == 'sqlite':
            return SQLiteConnector()

        raise UnsupportedDriver(driver)

    def _create_connection(self, driver, connection, database, prefix='', config=None):
        if config is None:
            config = {}

        if driver == 'mysql':
            return MySqlConnection(connection, database, prefix, config)
        elif driver == 'postgres':
            return PostgresConnection(connection, database, prefix, config)
        elif driver == 'sqlite':
            return SQLiteConnection(connection, database, prefix, config)

        raise UnsupportedDriver(driver)



# -*- coding: utf-8 -*-

import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase
from ...utils import MockConnection

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.processors import QueryProcessor
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import HasManyThrough
from eloquent.orm.collection import Collection


class OrmHasManyThroughTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_relation_is_properly_initialized(self):
        relation = self._get_relation()
        model = flexmock(Model())
        relation.get_related().should_receive('new_collection').replace_with(lambda l=None: Collection(l or []))
        model.should_receive('set_relation').once().with_args('foo', Collection)
        models = relation.init_relation([model], 'foo')

        self.assertEqual([model], models)

    def test_eager_constraints_are_properly_added(self):
        relation = self._get_relation()
        relation.get_query().get_query().should_receive('where_in').once().with_args('users.country_id', [1, 2])
        model1 = OrmHasManyThroughModelStub()
        model1.id = 1
        model2 = OrmHasManyThroughModelStub()
        model2.id = 2
        relation.add_eager_constraints([model1, model2])

    def test_models_are_properly_matched_to_parents(self):
        relation = self._get_relation()

        result1 = OrmHasManyThroughModelStub()
        result1.country_id = 1
        result2 = OrmHasManyThroughModelStub()
        result2.country_id = 2
        result3 = OrmHasManyThroughModelStub()
        result3.country_id = 2

        model1 = OrmHasManyThroughModelStub()
        model1.id = 1
        model2 = OrmHasManyThroughModelStub()
        model2.id = 2
        model3 = OrmHasManyThroughModelStub()
        model3.id = 3

        relation.get_related().should_receive('new_collection').replace_with(lambda l=None: Collection(l or []))
        models = relation.match([model1, model2, model3], Collection([result1, result2, result3]), 'foo')

        self.assertEqual(1, models[0].foo[0].country_id)
        self.assertEqual(1, len(models[0].foo))
        self.assertEqual(2, models[1].foo[0].country_id)
        self.assertEqual(2, models[1].foo[1].country_id)
        self.assertEqual(2, len(models[1].foo))
        self.assertFalse(hasattr(models[2], 'foo'))

    def _get_relation(self):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.get_query().should_receive('join').once().with_args('users', 'users.id', '=', 'posts.user_id')
        builder.should_receive('where').with_args('users.country_id', '=', 1)
        country = flexmock(Model())
        country.should_receive('get_key').and_return(1)
        country.should_receive('get_foreign_key').and_return('country_id')
        user = flexmock(Model())
        user.should_receive('get_table').and_return('users')
        user.should_receive('get_qualified_key_name').and_return('users.id')
        post = flexmock(Model())
        post.should_receive('get_table').and_return('posts')
        builder.should_receive('get_model').and_return(post)

        user.should_receive('get_key').and_return(1)
        user.should_receive('get_created_at_column').and_return('created_at')
        user.should_receive('get_updated_at_column').and_return('updated_at')

        parent = flexmock(Model())
        parent.should_receive('get_attribute').with_args('id').and_return(1)
        parent.should_receive('get_created_at_column').and_return('created_at')
        parent.should_receive('get_updated_at_column').and_return('updated_at')
        parent.should_receive('new_query').and_return(builder)

        return HasManyThrough(builder, country, user, 'country_id', 'user_id')


class OrmHasManyThroughModelStub(Model):

    pass



# -*- coding: utf-8 -*-


import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase

from eloquent.query.builder import QueryBuilder
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import HasOne


class OrmRelationTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_set_relation_fail(self):
        parent = OrmRelationResetModelStub()
        relation = OrmRelationResetModelStub()
        parent.set_relation('test', relation)
        parent.set_relation('foo', 'bar')
        self.assertFalse('foo' in parent.to_dict())

    def test_touch_method_updates_related_timestamps(self):
        builder = flexmock(Builder, get_model=None, where=None)
        parent = Model()
        parent = flexmock(parent)
        parent.should_receive('get_attribute').with_args('id').and_return(1)
        related = Model()
        related = flexmock(related)
        builder.should_receive('get_model').and_return(related)
        builder.should_receive('where')
        relation = HasOne(Builder(QueryBuilder(None, None, None)), parent, 'foreign_key', 'id')
        related.should_receive('get_table').and_return('table')
        related.should_receive('get_updated_at_column').and_return('updated_at')
        now = arrow.get()
        related.should_receive('fresh_timestamp').and_return(now)
        builder.should_receive('update').once().with_args({'updated_at': now})

        relation.touch()


class OrmRelationResetModelStub(Model):

    def get_query(self):
        return self.new_query().get_query()



# -*- coding: utf-8 -*-


import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import HasMany
from eloquent.orm.collection import Collection


class OrmHasManyTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_create_properly_creates_new_model(self):
        relation = self._get_relation()
        created = flexmock(Model(), save=lambda: True, set_attribute=lambda: None)
        created.should_receive('save').once().and_return(True)
        relation.get_related().should_receive('new_instance').once().with_args({'name': 'john'}).and_return(created)
        created.should_receive('set_attribute').with_args('foreign_key', 1)

        self.assertEqual(created, relation.create(name='john'))

    def test_find_or_new_finds_model(self):
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('find').once().with_args('foo', ['*']).and_return(model)
        model.should_receive('set_attribute').never()

        self.assertEqual('bar', relation.find_or_new('foo').foo)

    def test_find_or_new_returns_new_model_with_foreign_key_set(self):
        relation = self._get_relation()
        relation.get_query().should_receive('find').once().with_args('foo', ['*']).and_return(None)
        model = flexmock()
        model.foo = 'bar'
        relation.get_related().should_receive('new_instance').once().with_args().and_return(model)
        model.should_receive('set_attribute').once().with_args('foreign_key', 1)

        self.assertEqual('bar', relation.find_or_new('foo').foo)

    def test_first_or_new_finds_first_model(self):
        relation = self._get_relation()
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('first').once().with_args().and_return(model)
        model.should_receive('set_attribute').never()

        self.assertEqual('bar', relation.first_or_new(foo='bar').foo)

    def test_first_or_new_returns_new_model_with_foreign_key_set(self):
        relation = self._get_relation()
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(None)

        model = flexmock()
        model.foo = 'bar'
        relation.get_related().should_receive('new_instance').once().with_args().and_return(model)
        model.should_receive('set_attribute').once().with_args('foreign_key', 1)

        self.assertEqual('bar', relation.first_or_new(foo='bar').foo)

    def test_first_or_create_finds_first_model(self):
        relation = self._get_relation()
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('first').once().with_args().and_return(model)
        model.should_receive('set_attribute').never()

        self.assertEqual('bar', relation.first_or_create(foo='bar').foo)

    def test_first_or_create_returns_new_model_with_foreign_key_set(self):
        relation = self._get_relation()
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(None)

        model = flexmock()
        model.foo = 'bar'
        relation.get_related().should_receive('new_instance').once().with_args({'foo': 'bar'}).and_return(model)
        model.should_receive('save').once().and_return(True)
        model.should_receive('set_attribute').once().with_args('foreign_key', 1)

        self.assertEqual('bar', relation.first_or_create(foo='bar').foo)

    def test_update_or_create_finds_first_model_and_updates(self):
        relation = self._get_relation()
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())

        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('first').once().with_args().and_return(model)
        relation.get_related().should_receive('new_instance').never()
        model.should_receive('fill').once().with_args({'foo': 'baz'})
        model.should_receive('save').once()

        self.assertEqual('bar', relation.update_or_create({'foo': 'bar'}, {'foo': 'baz'}).foo)

    def test_update_or_create_creates_new_model_with_foreign_key_set(self):
        relation = self._get_relation()
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())

        relation.get_query().should_receive('first').once().with_args().and_return(None)

        model = flexmock()
        model.foo = 'bar'
        relation.get_related().should_receive('new_instance').once().and_return(model)
        model.should_receive('fill').once().with_args({'foo': 'baz'})
        model.should_receive('save').once()
        model.should_receive('set_attribute').once().with_args('foreign_key', 1)

        self.assertEqual('bar', relation.update_or_create({'foo': 'bar'}, {'foo': 'baz'}).foo)

    def test_update_updates_models_with_timestamps(self):
        relation = self._get_relation()
        relation.get_related().should_receive('uses_timestamps').once().and_return(True)
        now = arrow.get()
        relation.get_related().should_receive('fresh_timestamp').once().and_return(now)
        relation.get_query().should_receive('update').once().with_args({'foo': 'bar', 'updated_at': now}).and_return('results')

        self.assertEqual('results', relation.update(foo='bar'))

    def test_relation_is_properly_initialized(self):
        relation = self._get_relation()
        model = flexmock(Model())
        model.should_receive('set_relation').once().with_args('foo', Collection)
        models = relation.init_relation([model], 'foo')

        self.assertEqual([model], models)

    def test_eager_constraints_are_properly_added(self):
        relation = self._get_relation()
        relation.get_query().get_query().should_receive('where_in').once().with_args('table.foreign_key', [1, 2])

        model1 = OrmHasOneModelStub()
        model1.id = 1
        model2 = OrmHasOneModelStub()
        model2.id = 2

        relation.add_eager_constraints([model1, model2])

    def test_models_are_properly_matched_to_parents(self):
        relation = self._get_relation()

        result1 = OrmHasOneModelStub()
        result1.foreign_key = 1
        result2 = OrmHasOneModelStub()
        result2.foreign_key = 2
        result3 = OrmHasOneModelStub()
        result3.foreign_key = 2

        model1 = OrmHasOneModelStub()
        model1.id = 1
        model2 = OrmHasOneModelStub()
        model2.id = 2
        model3 = OrmHasOneModelStub()
        model3.id = 3

        relation.get_related().should_receive('new_collection').replace_with(lambda l: Collection(l))

        models = relation.match([model1, model2, model3], Collection([result1, result2, result3]), 'foo')

        self.assertEqual(1, models[0].foo[0].foreign_key)
        self.assertEqual(1, len(models[0].foo))
        self.assertEqual(2, models[1].foo[0].foreign_key)
        self.assertEqual(2, models[1].foo[1].foreign_key)
        self.assertEqual(2, len(models[1].foo))
        self.assertFalse(hasattr(models[2], 'foo'))

    def test_relation_count_query_can_be_built(self):
        relation = self._get_relation()
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.get_query().should_receive('select').once()
        relation.get_parent().should_receive('get_table').and_return('table')
        builder.should_receive('where').once().with_args('table.foreign_key', '=', QueryExpression)
        parent_query = flexmock(QueryBuilder(None, None, None))
        relation.get_query().should_receive('get_query').and_return(parent_query)
        grammar = flexmock()
        parent_query.should_receive('get_grammar').once().and_return(grammar)
        grammar.should_receive('wrap').once().with_args('table.id')

        relation.get_relation_count_query(builder, builder)

    def _get_relation(self):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.should_receive('where').with_args('table.foreign_key', '=', 1)
        related = flexmock(Model())
        builder.should_receive('get_model').and_return(related)
        parent = flexmock(Model())
        parent.should_receive('get_attribute').with_args('id').and_return(1)
        parent.should_receive('get_created_at_column').and_return('created_at')
        parent.should_receive('get_updated_at_column').and_return('updated_at')
        parent.should_receive('new_query').and_return(builder)

        return HasMany(builder, parent, 'table.foreign_key', 'id')


class OrmHasOneModelStub(Model):

    pass



# -*- coding: utf-8 -*-


class Collection(object):

    def __init__(self, items=None):
        """
        Creates a new Collection

        :param items: The collection items
        :type items: dict or list or Collection

        :rtype: None
        """
        if items is None:
            items = []
        else:
            items = self._get_items(items)

        if not isinstance(items, (list, dict, tuple)):
            self._items = [items]
        else:
            self._items = items

    @classmethod
    def make(cls, items=None):
        """
        Create a new Collection instance if the value isn't one already

        :param items: The collection items
        :type items: dict or list or Collection

        :return: A Collection instance
        :rtype: Collection
        """
        if isinstance(items, Collection):
            return items

        return cls(items)

    def all(self):
        """
        Get all of the items in the collection

        :return: The items in the collections
        :type: mixed
        """
        return self._items

    def collapse(self):
        """
        Collapse the collection items into a single element (dict or list)

        :return: A new Collection instance with collapsed items
        :rtype: Collection
        """
        results = []

        if isinstance(self._items, dict):
            items = self._items.values()

        for values in items:
            if isinstance(values, Collection):
                values = values.all()

            results += values

        return Collection(results)

    def contains(self, key, value=None):
        """
        Determine if an element is in the collection

        :param key: The element
        :type key: int or str

        :param value: The value of the element
        :type value: mixed

        :return: Whether the element is in the collection
        :rtype: bool
        """
        if value is not None:
            if isinstance(self._items, list):
                return key in self._items and self._items[self._items.index(key)] == value

            return self._items.get(key) == value

        return key in self._items

    def __contains__(self, item):
        return self.contains(item)

    def diff(self, items):
        """
        Diff the collections with the given items

        :param items: The items to diff with
        :type items: mixed

        :return: A Collection instance
        :rtype: Collection
        """
        pass

    def first(self, default=None):
        """
        Get the first item of the collection.

        :param default: The default value
        :type default: mixed
        """
        if len(self._items) > 0:
            return self._items[0]
        else:
            return default

    def lists(self, value, key=None):
        """
        Get a list with the values of a given key

        :rtype: list
        """
        results = map(lambda x: x[value], self._items)

        return list(results)

    def map(self, callback):
        """
        Run a map over each of the item.

        :param callback: The map function
        :type callback: callable

        :rtype: Collection
        """
        if isinstance(self._items, dict):
            return Collection(list(map(callback, self._items.values())))

        return Collection(list(map(callback, self._items)))

    def unique(self):
        """
        Return only unique items from the collection list.

        :rtype: Collection
        """
        seen = set()
        seen_add = seen.add

        return Collection([x for x in self._items if not (x in seen or seen_add(x))])

    def _get_items(self, items):
        if isinstance(items, Collection):
            items = items.all()
        elif hasattr('items', 'to_list'):
            items = items.to_list()
        elif hasattr('items', 'to_dict'):
            items = items.to_dict()

        return items

    def to_dict(self):
        return list(map(lambda value: value.to_dict() if hasattr(value, 'to_dict') else value,
                        self._items))

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        for item in self._items:
            yield item

    def __getitem__(self, item):
        return self._items[item]



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

__version__ = '0.3.1'

setup(
    name='eloquent',
    license='MIT',
    version=__version__,
    description='The Eloquent ORM provides a simple yet beautiful ActiveRecord implementation.',
    long_description=open('README.rst').read(),
    author='Sébastien Eustace',
    author_email='sebastien.eustace@gmail.com',
    url='https://github.com/sdispater/eloquent',
    download_url='https://github.com/sdispater/eloquent/archive/%s.tar.gz' % __version__,
    packages=find_packages(),
    install_requires=['simplejson', 'arrow', 'inflection', 'six'],
    tests_require=['pytest', 'mock', 'flexmock'],
    test_suite='nose.collector',
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-

from .processor import QueryProcessor


class MySqlQueryProcessor(QueryProcessor):

    def process_insert_get_id(self, query, sql, values, sequence=None):
        """
        Process an "insert get ID" query.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param sql: The sql query to execute
        :type sql: str

        :param values: The value bindings
        :type values: list

        :param sequence: The ids sequence
        :type sequence: str

        :return: The inserted row id
        :rtype: int
        """
        if not query.get_connection().transaction_level():
            with query.get_connection().transaction():
                query.get_connection().insert(sql, values)

                cursor = query.get_connection().get_cursor()
                if hasattr(cursor, 'lastrowid'):
                    id = cursor.lastrowid
                else:
                    id = query.get_connection().statement('SELECT LAST_INSERT_ID()')
        else:
            query.get_connection().insert(sql, values)

            cursor = query.get_connection().get_cursor()
            if hasattr(cursor, 'lastrowid'):
                id = cursor.lastrowid
            else:
                id = query.get_connection().statement('SELECT LAST_INSERT_ID()')

        if isinstance(id, int):
            return id

        if str(id).isdigit():
            return int(id)

        return id

    def process_column_listing(self, results):
        """
        Process the results of a column listing query

        :param results: The query results
        :type results: dict

        :return: The processed results
        :return: dict
        """
        return map(lambda x: x['column_name'], results)



# -*- coding: utf-8 -*-

import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import MorphTo
from eloquent.orm.collection import Collection


class OrmMorphToTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_lookup_dictionary_is_properly_constructed(self):
        relation = self._get_relation()

        one = flexmock()
        one.morph_type = 'morph_type_1'
        one.foreign_key = 'foreign_key_1'
        two = flexmock()
        two.morph_type = 'morph_type_1'
        two.foreign_key = 'foreign_key_1'
        three = flexmock()
        three.morph_type = 'morph_type_2'
        three.foreign_key = 'foreign_key_2'

        relation.add_eager_constraints([one, two, three])

        dictionary = relation.get_dictionary()

        self.assertEqual({
            'morph_type_1': {
                'foreign_key_1': [
                    one,
                    two
                ]
            },
            'morph_type_2': {
                'foreign_key_2': [three]
            }
        }, dictionary)

    def test_models_are_properly_pulled_and_matched(self):
        relation = self._get_relation()

        one = flexmock(Model())
        one.morph_type = 'morph_type_1'
        one.foreign_key = 'foreign_key_1'
        two = flexmock(Model())
        two.morph_type = 'morph_type_1'
        two.foreign_key = 'foreign_key_1'
        three = flexmock(Model())
        three.morph_type = 'morph_type_2'
        three.foreign_key = 'foreign_key_2'

        relation.add_eager_constraints([one, two, three])

        first_query = flexmock(Builder(flexmock(QueryBuilder(None, QueryGrammar(), None))))
        second_query = flexmock(Builder(flexmock(QueryBuilder(None, QueryGrammar(), None))))
        first_model = flexmock(Model())
        second_model = flexmock(Model())
        relation.should_receive('_create_model_by_type').once().with_args('morph_type_1').and_return(first_model)
        relation.should_receive('_create_model_by_type').once().with_args('morph_type_2').and_return(second_model)
        first_model.should_receive('get_key_name').and_return('id')
        second_model.should_receive('get_key_name').and_return('id')

        first_model.should_receive('new_query').once().and_return(first_query)
        second_model.should_receive('new_query').once().and_return(second_query)

        first_query.get_query().should_receive('where_in').once()\
            .with_args('id', ['foreign_key_1']).and_return(first_query)
        result_one = flexmock()
        first_query.should_receive('get').and_return(Collection.make([result_one]))
        result_one.should_receive('get_key').and_return('foreign_key_1')

        second_query.get_query().should_receive('where_in').once()\
            .with_args('id', ['foreign_key_2']).and_return(second_query)
        result_two = flexmock()
        second_query.should_receive('get').and_return(Collection.make([result_two]))
        result_two.should_receive('get_key').and_return('foreign_key_2')

        one.should_receive('set_relation').once().with_args('relation', result_one)
        two.should_receive('set_relation').once().with_args('relation', result_one)
        three.should_receive('set_relation').once().with_args('relation', result_two)

        relation.get_eager()

    # TODO: soft deletes

    def test_associate_sets_foreign_key_and_type_on_model(self):
        parent = flexmock(Model())
        parent.should_receive('get_attribute').once().with_args('foreign_key').and_return('foreign.value')

        relation = self._get_relation_associate(parent)

        associate = flexmock(Model())
        associate.should_receive('get_key').once().and_return(1)
        associate.should_receive('get_morph_class').once().and_return('Model')

        parent.should_receive('set_attribute').once().with_args('foreign_key', 1)
        parent.should_receive('set_attribute').once().with_args('morph_type', 'Model')
        parent.should_receive('set_relation').once().with_args('relation', associate)

        relation.associate(associate)

    def _get_relation_associate(self, parent):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.should_receive('where').with_args('relation.id', '=', 'foreign.value')
        related = flexmock(Model())
        related.should_receive('get_key').and_return(1)
        related.should_receive('get_table').and_return('relation')
        builder.should_receive('get_model').and_return(related)

        return MorphTo(builder, parent, 'foreign_key', 'id', 'morph_type', 'relation')

    def _get_relation(self, parent=None, builder=None):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = builder or Builder(query)
        builder.should_receive('where').with_args('relation.id', '=', 'foreign.value')
        related = flexmock(Model())
        related.should_receive('get_key').and_return(1)
        related.should_receive('get_table').and_return('relation')
        builder.should_receive('get_model').and_return(related)

        parent = parent or OrmMorphToModelStub()

        flexmock(MorphTo)
        return MorphTo(builder, parent, 'foreign_key', 'id', 'morph_type', 'relation')


class OrmMorphToModelStub(Model):

    foreign_key = 'foreign.value'
